#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""
Functions for exporting into the text-only human-readable
`YAFF format <https://github.com/robhagemans/monobit/blob/master/YAFF.md>`_.
"""
from __future__ import annotations

import mcfonts.glyphs
import mcfonts.utils.unihex

YAFF_STORE_GLYPH_MAGIC = "__glyph__"
"""The magic constant that is used as a key for glyph data to be stored under."""


def set_space_to_character(
    yaff_store: dict[str, str | int | dict[str, list[str]]],
    character: str,
    width: int,
    replace: bool = True,
) -> None:
    """
    Set a simple blank space width to a character.

    :param yaff_store: The YAFF font, represented as a dictionary.
    :param character: The character to set a space to.
    :param width: The width.
    :param replace: Whether the width should overwrite pre-existing glyph data for `character`.
    """
    local_name = f"u+{ord(character):04X}"
    if local_name in yaff_store and not replace:
        return
    if local_name not in yaff_store:
        yaff_store[local_name] = {YAFF_STORE_GLYPH_MAGIC: []}
    yaff_store[local_name][YAFF_STORE_GLYPH_MAGIC] = ["." * width] * 8


def set_yaff_to_character(
    yaff_store: dict[str, str | int | dict[str, list[str] | str | int]],
    yaff_glyph: dict[str, list[str] | str | int],
    character: str,
    replace: bool = True,
) -> None:
    """
    Set YAFF glyph data to a character.

    :param yaff_store: The YAFF font, represented as a dictionary.
    :param yaff_glyph: A dictionary of the YAFF data. Glyph data is stored under :data:`YAFF_STORE_GLYPH_MAGIC`.
    :param character: The character to set a glyph to.
    :param replace: Whether the width should overwrite pre-existing glyph data for `character`.
    """
    if (local_name := f"u+{ord(character):04X}") in yaff_store and not replace:
        return
    yaff_store[local_name] = yaff_glyph


def include_glyph(
    yaff_store: dict[str, str | int | dict[str, list[str]]],
    glyph: mcfonts.glyphs.Glyph,
    character: str,
) -> None:
    """
    Include `glyph` into the YAFF font, represented as a dictionary.

    Different Glyph subtypes have different methods of being imported;
    this function centralizes those varying methods.

    :param yaff_store: The YAFF font, represented as a dictionary.
    :param glyph: An instance of :class:`mcfont.glyphs.Glyph`.
    :param character: The character to assign this glyph to.
    """
    if isinstance(glyph, mcfonts.glyphs.SpaceGlyph):
        set_space_to_character(yaff_store, character, min(glyph.get_width(), 512))
    elif isinstance(glyph, mcfonts.glyphs.BitmapGlyph):
        set_yaff_to_character(yaff_store, generate_bitmap_yaff(glyph), character, False)
    elif isinstance(glyph, mcfonts.glyphs.UnihexGlyph):
        set_yaff_to_character(yaff_store, generate_unihex_yaff(glyph), character, False)


def generate_bitmap_yaff(glyph: mcfonts.glyphs.BitmapGlyph) -> dict[str, list[str] | str | int]:
    """
    Generate a YAFF glyph out of a :class:`~mcfonts.glyphs.BitmapGlyph`.

    :param glyph: A :class:`~mcfonts.glyphs.BitmapGlyph`.
    :returns: A dictionary of the YAFF data. Glyph data is stored under :data:`YAFF_STORE_GLYPH_MAGIC`.
    """

    image = glyph.get_image()
    pixel_values = [pixel >= mcfonts.glyphs.BITMAP_EXPORT_PIXEL_THRESHOLD for pixel in image.convert("LA").getdata(1)]

    glyph_store = []

    for row in [pixel_values[i : i + image.width] for i in range(0, len(pixel_values), image.width)]:
        # Trim the excess to the real width.
        # There'll never be data there anyway.
        glyph_store.append(
            "".join("@" if pixel_value else "." for pixel_value in row)[
                : mcfonts.utils.image.get_image_bearings(image)[1]
            ]
            + "."  # Single padding
        )

    return {
        YAFF_STORE_GLYPH_MAGIC: glyph_store,
        "ascent": glyph.get_ascent(),
        "descent": glyph.get_height() - glyph.get_ascent(),
        "pixel-size": glyph.get_height() + glyph.get_ascent(),
    }


def generate_unihex_yaff(glyph: mcfonts.glyphs.UnihexGlyph) -> dict[str, list[str] | str | int]:
    """
    Generate a YAFF glyph out of a :class:`~mcfonts.glyphs.UnihexGlyph`.

    :param glyph: A :class:`~mcfonts.glyphs.UnihexGlyph`.
    :returns: A dictionary of the YAFF data. Glyph data is stored under :data:`YAFF_STORE_GLYPH_MAGIC`.
    """
    return {
        YAFF_STORE_GLYPH_MAGIC: [
            row.replace("0", ".").replace("1", "@")
            for row in mcfonts.utils.unihex.bit_string_to_rows(glyph.get_bit_string(), False)
        ],
        "ascent": 7,
        "descent": 1,
    }



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""
Functions for exporting into the GNU Unifont HEX format.

Contains functions for transforming and handling Unifont HEX data.
"""
from __future__ import annotations

import PIL.Image

import mcfonts.glyphs


def set_space_to_character(hex_store: dict[str, str], character: str, width: int, replace: bool = True) -> None:
    """
    Set a simple blank space width to a character.

    :param hex_store: The .hex font, represented as a dictionary.
    :param character: The character to set a space to.
    :param width: The width.
    :param replace: Whether the width should overwrite pre-existing glyph data for `character`.
    """
    local_name = f"{ord(character):04X}"
    if local_name in hex_store and not replace:
        return
    hex_store[local_name] = "0" * (width * 4)


def set_hex_to_character(hex_store: dict[str, str], hex_glyph: str, character: str, replace: bool = True) -> None:
    """
    Set .hex glyph data to a character.

    :param hex_store: The .hex font, represented as a dictionary.
    :param hex_glyph: A string of the glyph data.
    :param character: The character to set a glyph to.
    :param replace: Whether the width should overwrite pre-existing glyph data for `character`.
    """
    local_name = f"{ord(character):04X}"
    if local_name in hex_store and not replace:
        return
    hex_store[local_name] = hex_glyph


def generate_bitmap_hex(glyph: mcfonts.glyphs.BitmapGlyph) -> str:
    """
    Generate a .hex glyph out of a :class:`~mcfonts.glyphs.BitmapGlyph`.

    :param glyph: A :class:`~mcfonts.glyphs.BitmapGlyph`.
    :returns: A string of the .hex glyph data. In hexadecimal, used as-is for exporting.
    """
    if glyph.get_bearings() == (0, 0):
        return "0000"

    image = glyph.get_image()
    glyph_width = image.width

    newsize_image = image.copy()
    # Image must be 16 pixels tall and 2**n height to be encoded as unihex
    if image.height > 16 or image.width > 16:  # We can just resize image down to force it to fit
        newsize_image = image.resize((16, 2 ** (glyph_width - 1).bit_length()), PIL.Image.Resampling.NEAREST)
    elif image.height != 16 or (glyph_width & (glyph_width - 1) != 0):
        newsize_image = PIL.Image.new("LA", (16, 2 ** (glyph_width - 1).bit_length()))
        newsize_image.paste(image, (16 + (glyph.get_height() - glyph.get_ascent()), 0))

    data = newsize_image.convert("LA").getdata(1)

    bit_string = []
    row = 0
    for index, pixel in enumerate(data):
        if pixel >= mcfonts.glyphs.BITMAP_EXPORT_PIXEL_THRESHOLD:
            # Set the bit in `row` at position equal to the X position in the current bitmap image row
            row |= 1 << (glyph_width - (index % glyph_width))
        if index % (glyph_width - 1) == 0:
            # Index position is at the end of glyph X boundary, start new row
            # Index starts at 0, and width starts at 1
            # Pad "byte" with 0 to match the number of bytes it would be, equal to width // 4 (16 width,
            bit_string.append(hex(row)[2:].rjust(glyph_width // 4, "0"))
            row = 0

    return "".join(bit_string)


def include_glyph(hex_store: dict[str, str], glyph: mcfonts.glyphs.Glyph, character: str) -> None:
    """
    Include `glyph` into the .hex font, represented as a dictionary.

    Different Glyph subtypes have different methods of being imported;
    this function centralizes those varying methods.

    :param hex_store: The .hex font, represented as a dictionary.
    :param glyph: An instance of :class:`mcfont.glyphs.Glyph`.
    :param character: The character to assign this glyph to.
    """
    if isinstance(glyph, mcfonts.glyphs.SpaceGlyph):
        set_space_to_character(hex_store, character, min(glyph.get_width(), 512))
    elif isinstance(glyph, mcfonts.glyphs.BitmapGlyph):
        set_hex_to_character(hex_store, generate_bitmap_hex(glyph), character, False)
    elif isinstance(glyph, mcfonts.glyphs.UnihexGlyph):
        set_hex_to_character(hex_store, glyph.get_bit_string(), character, False)



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""Contains the basic parent provider that all functional providers subclass and inherit from."""
from __future__ import annotations

import abc
import collections.abc
import os
import typing

import mcfonts.constants
import mcfonts.glyphs


class Provider(abc.ABC):
    """The base provider class. All providers subclass this."""

    def __init__(
        self,
        provider_type: str,
        origin: os.PathLike[str] | str,
        chars_covered: set[str] | None = None,
    ):
        """
        Construct a Provider base class.

        :param provider_type: A string of the value in the JSON "type" field.
        :param origin: The absolute path to the provider's JSON file.
        :param chars_covered: A set of the individual characters covered by this provider.
        """

        if chars_covered is None:
            chars_covered = set()
        self.provider_type = provider_type
        self.chars_covered = chars_covered
        self.origin = origin

    def __repr__(self) -> str:
        return (
            f'{self.__class__.__name__}("{self.provider_type}", {self.get_contents()}, '
            f"{self.get_pack_versions()[0]},"
            f" {self.chars_covered})"
        )

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({len(self.chars_covered):,} chars)"

    def validate(self) -> None:
        """
        Specific tests for validating a provider.

        This function checks values to ensure that they're correct and in an acceptable range.

        If applicable, textures are checked to ensure all characters have a defined glyph,
        and that the dimensions are correct.

        Warnings are printed through :func:`mcfonts.logger.warning`.

        :returns: Nothing; problems are either raised or warned.
        """

    def print_info(self) -> None:
        """Print information about this provider."""
        print(f"Provider type: {mcfonts.constants.UNKNOWN_FIELD}")
        print(f"Characters covered: {len(self.chars_covered):,}")

    @abc.abstractmethod
    def pretty_print(self) -> str:
        """
        Return a short one line description of the provider.

        :returns: A string of the provider's info, normally ``<icon> <type>: <info,...>``.
        """
        return f"{self.get_icon()} {self.provider_type}"

    @abc.abstractmethod
    def construct_glyphs(self) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.Glyph]]:
        """
        Yield a tuple of (character, glyph) for every glyph available in the provider's resource.

        Some glyphs may be empty or have no content.

        :returns: A yield of ``(character, Glyph)``, or None if there are no glyphs.
        """

    @abc.abstractmethod
    def yield_characters(self) -> collections.abc.Iterable[str]:
        """
        Yield strings of the individual characters this provider supports.

        :returns: A yield of strings of length 1, or None if no characters are present.
        """

    @abc.abstractmethod
    def yield_glyphs(self) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.Glyph | None]]:
        """
        Yield a tuple of (character, glyph) for every glyph available in the provider's resource.

        This is different from :meth:`construct_glyphs`
        because it yields glyphs that have already been constructed and stored.

        :returns: A yield of ``(character, Glyph)``.
        """

    @abc.abstractmethod
    def get_icon(self) -> str:
        """
        Return a Unicode character that can be used to represent this provider in shorthand descriptions.

        :returns: A 1-character string.
        """
        return "?"

    @abc.abstractmethod
    def get_pack_versions(self) -> tuple[int, int | None]:
        """
        Return a tuple of the minimum and maximum pack versions for which this provider is valid.

        The first element is guaranteed to be an integer. There will always be a minimum version, even if it's 0.
        The second element may be an integer, or None. If it is None, there is no maximum version bound.

        :returns: A 2-element tuple of the minimum and maximum valid pack versions.
        """
        return 0, None

    @abc.abstractmethod
    def yield_glyphs_in_unirange(
        self, unirange_notation: str
    ) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.Glyph | None]]:
        """
        Given `unirange_notation`, return a dictionary of the requested characters to their glyphs.

        :param unirange_notation:
            A string representing the requested range of chars.
            See https://pypi.org/project/unirange/.
        :returns:
            A yield of the requested glyphs that match `unirange_notation`.
        """

    @abc.abstractmethod
    def get_contents(self) -> mcfonts.utils.types.TypedProviderDict:
        """
        Return the contents of the provider.

        :returns: Dictionary of the provider.
        """


def format_provider_message(provider: mcfonts.providers.base.Provider | dict[str, typing.Any], message: str) -> str:
    """
    Format a provider warning message properly, following Provider <type>: <message>.

    Calls :func:`mcfonts.utils.pretty_print_provider()`.

    :param provider:
        The provider, either as a dictionary or instance of :data:`mcfonts.providers.base.Provider`.
    :param message: The message to append at the end.
    :returns: The formatted warning message.
    """
    if isinstance(provider, dict):
        return f"Provider {pretty_print_provider_dictionary(provider)} {message}"
    return f"Provider {provider.pretty_print()} {message}"


def pretty_print_provider_dictionary(provider: dict[str, typing.Any]) -> str:
    """
    Format a provider information message properly, following ``<type><provider specific>``.

    Provider specific info:

     * ``bitmap``: ``<file> h <height> a <ascent>``
     * ``space``: nothing
     * ``legacy_unicode``: ``<template>``
     * ``ttf``: ``<file> s <shift0, 1>, sz <size>, o <oversample>, sk <skip>``

    :param provider:
        The provider as a dictionary, **not** an instance of :data:`mcfonts.providers.base.Provider`.
    :returns: The pretty provider information.
    """
    if (provider_type := provider.get("type", "")).lower() == "bitmap":
        return f'"bitmap": {provider.get("file", "no resource")}'
    if provider_type == "space":
        return '"space"'
    if provider_type == "legacy_unicode":
        return f'"legacy_unicode": {provider.get("template", "no template")}'
    if provider_type == "ttf":
        return (
            f'"ttf": {provider.get("file", "no file")}, s '
            f'{provider.get("shift", "[?, ?]")}, sz '
            f'{provider.get("size", "?")}, o '
            f'{provider.get("oversample", "?")}, sk '
            f'{provider.get("skip", "none")}'
        )
    if provider_type == "reference":
        return f'"reference": to {provider.get("id"), "nothing"}'
    if provider_type == "unihex":
        return f'"unihex": {provider.get("hex_file")}, so {len(provider.get("size_overrides", []))}'
    return f'"{provider_type}": ?'



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""A unihex provider, handling .hex glyphs."""
import collections.abc
import os
import typing

import jsonschema
import unirange

import mcfonts.glyphs
import mcfonts.providers.base
import mcfonts.utils.schemas
import mcfonts.utils.unihex


class UnihexProvider(mcfonts.providers.base.Provider):
    """
    The ``unihex`` provider loads glyphs from text files in the ".hex" format.

    It has two required fields, ``hex_file`` and ``size_overrides``.

    * ``hex_file`` is a resource location to a .ZIP file containing any number of ".hex" files at its root.
    * ``size_overrides`` is a list of maps stating the beginning and end of a character range, and the starting and
      ending columns of the glyphs that match that range.

    Each ".hex" file is a text document with each line containing a hex glyph.
    A hex glyph is a string of text in the format of ``<codepoint>:<bit string>``.
    The bit string is a string of hexadecimal numbers.
    The numbers, when converted to binary, describe the glyph's pixels.
    A 1 bit is an "on" pixel, and a 0 bit is an "off" pixel.
    When exporting, the empty left and right columns are trimmed unless ``size_overrides`` explicitly overrides this.
    """

    def __init__(
        self,
        provider: mcfonts.utils.types.TypedProviderUnihex,
        origin: os.PathLike[str] | str,
        resources: dict[str, str] | None,
    ) -> None:
        self.contents = provider
        self.resources = resources
        self.size_overrides = dict(self.yield_size_overrides())
        self.glyphs = dict(self.construct_glyphs())
        super().__init__("unihex", origin, set(self.glyphs.keys()))
        self.validate()

    def construct_glyphs(self) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.UnihexGlyph]]:
        if not self.resources:
            return
        for name, hex_contents in self.resources.items():
            for hex_glyph in hex_contents.splitlines():
                yield mcfonts.utils.unihex.get_codepoint(hex_glyph), mcfonts.glyphs.UnihexGlyph(
                    mcfonts.utils.unihex.get_bit_string(hex_glyph), None, None, None, None
                )

    def yield_size_overrides(self) -> collections.abc.Iterable[tuple[str, tuple[int, int]]]:
        """
        Yield all the size overrides for every character.

        Overrides are given as ``tuple[int, int]``.

        :returns: A yield of ``(character: str, (left: int, right: int))``.
        """
        for override in self.get_contents().get("size_overrides", {}):
            left = override.get("left", -1)
            right = override.get("right", -1)
            for codepoint in range(ord(override.get("from", "\0")), ord(override.get("to", "\0"))):
                yield chr(codepoint), (left, right)

    def get_size_override(self, character: str) -> tuple[int, int] | None:
        """
        Return the size overrides of a character, if they exist.

        Size overrides is a tuple of two integers of the starting and ending columns to cut off when exporting.
        If there are no size overrides for a character, None is returned.

        :returns: A tuple of two integers, or None.
        """
        return self.size_overrides.get(character)

    def validate(self) -> None:
        jsonschema.validate(self.contents, mcfonts.utils.schemas.SCHEMA_PROVIDER_UNIHEX)

    def pretty_print(self) -> str:
        return (
            f"{self.get_icon()} unihex: {len(self.resources or []):,} .hex files, "
            f"{len(self.size_overrides):,} size overrides"
        )

    def print_info(self) -> None:
        pass

    def yield_glyphs(self) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.UnihexGlyph]]:
        yield from self.glyphs.items()

    def yield_characters(self) -> collections.abc.Iterable[str]:
        yield from self.glyphs.keys()

    def get_icon(self) -> str:
        return "⬣"

    def get_pack_versions(self) -> tuple[int, None]:
        return 15, None

    def yield_glyphs_in_unirange(
        self, unirange_notation: str
    ) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.UnihexGlyph]]:
        for character in unirange.unirange_to_characters(unirange_notation):
            yield character, self.glyphs[character]  # Guaranteed to succeed, no need to dict.get().

    def get_contents(self) -> mcfonts.utils.types.TypedProviderUnihex:
        return self.contents



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""Functions for working with Unicode characters, codepoints, and surrogate pairs."""

from __future__ import annotations

import unicodedata

INVISIBLE_CHARACTERS = {
    "\u034F",
    "\u115F",
    "\u1160",
    "\u17B4",
    "\u17B5",
    "\u2800",
    "\u3164",
    "\uFFA0",
    "\U0001D159",
    "\U0001d173",
    "\U0001d174",
    "\U0001d175",
    "\U0001d176",
    "\U0001d177",
    "\U0001d178",
    "\U0001d179",
    "\U0001d17a",
}
"""A set of characters that do not have a visual representation under most fonts."""


def character_to_surrogates(character: str) -> tuple[int, int]:
    r"""
    Convert one character into 2 integers of its UTF-16 surrogate codepoints.

    A surrogate pair are two characters that represent another character.
    Since UTF-16 only stores characters from 0 to 0xFFFF,
    chars past 0xFFFF need to be split into two codepoints below 0xFFFF.

    This is useful even in plaintext Unicode notation, because ``\u1D105`` is not a single
    character, it's two (``ᴐ5``, not ``𝄅``).

    :param character:
        A single character.
    :returns:
        A surrogate pair, in codepoints of the surrogates.
    """
    codepoint = ord(character)
    codepoint -= 0x10000
    return 0xD800 + (codepoint >> 10), 0xDC00 + (codepoint & 0x3FF)


def surrogates_to_character(surrogates: tuple[int, int]) -> str:
    """
    Given a tuple of surrogate chars, return the single codepoint they combine to.

    :param surrogates:
        A tuple of two surrogate codepoints.
    :returns:
        A single character of the resulting surrogates.
    """
    return chr((surrogates[0] - 0xD800) * 0x400 + (surrogates[1] - 0xDC00) + 0x10000)


def is_character_invisible(character: str) -> bool:
    """
    Return if `character` would be invisible (might not have glyph info).

    A character is "invisible" if it:

    * Is in these categories: ``Cf, Cc, Zl, Zs, Zp``.
    * Is equal to these codepoints:
      ``2800, 034F, 115F, 1160, 17B4, 17B5, 3164, FFA0, 1D159, 1D174, 1D176, 1D177, 1D178, 1D17A``.
    * Is private use.

    You can visit `<https://invisible-characters.com/>`_ if you would like to see the list.

    .. warning::

        "Invisibility" is not a valid Unicode standard property.
        For standardization purposes, do use utilize this outside of this library.

    :param character:
        A single character.
    :returns:
        If `char` is a spacing character.
    """
    return (
        unicodedata.category(character) in ("Cf", "Cc", "Zl", "Zs", "Zp")
        or character in INVISIBLE_CHARACTERS
        or is_character_private_use(character)
    )


def is_character_private_use(character: str) -> bool:
    """
    Return if `character` is in a Private Use Area (PUA).

    A PUA is one of these codepoint ranges:

    * U+E000 to U+F8FF
    * U+F0000 to U+FFFFD
    * U+100000 to U+10FFFD

    :param character:
        A single character.
    :returns:
        If `character` is in a Private Use Area.
    """
    codepoint = ord(character)
    return (0xE000 <= codepoint <= 0xF8FF) or (0xF0000 <= codepoint <= 0xFFFFD) or (0x100000 <= codepoint <= 0x10FFFD)


def str_to_tags(string: str) -> str:
    """
    Return a version of `string` with all alphanumeric characters changed into Tags.

    Given `string`, which should have only ASCII characters, turn it into that same string but every character is a\
    Tag of itself, instead.

    See https://en.wikipedia.org/wiki/Tags_(Unicode_block).

    :param string: Any string; it should have ASCII characters.
    :returns: `string` but with ASCII characters replaced with their Tag equivalents.
    """
    return "".join(chr(ord(ch) + 0xE0000) for ch in string if ord(ch) < 0x7F)


def pretty_print_character(character: str) -> str:
    r"""
    Return relevant info about a character into a string, following ``U+<codepoint> <name> <character>``.

    >>> pretty_print_character('\u2601')
    'U+2601: CLOUD ☁'
    >>> pretty_print_character('\ue000')
    'U+E000: <PRIVATE USE> \ue000'
    >>> pretty_print_character('\U0001f400')
    'U+1F400: RAT 🐀'
    >>> pretty_print_character('\b')
    'U+0008: BACKSPACE ␈'
    >>> pretty_print_character('\b')
    'U+0008: BACKSPACE ␈'

    :param character: A single character.
    :returns: The pretty character string.
    """
    codepoint = ord(character)
    try:
        if unicodedata.category(character) == "Cc":
            return (
                f"U+{codepoint:04X}: "
                f"{unicodedata.name(chr(ord(character) + 0x2400)).split('SYMBOL FOR ')[1]} "
                f"{chr(ord(character) + 0x2400)}"
            )
        return f"U+{codepoint:04X}: {unicodedata.name(character)} {character}"
    except (ValueError, IndexError):
        if is_character_private_use(character):
            return f"U+{codepoint:04X}: <PRIVATE USE> {character}"
        return f"U+{codepoint:04X}: {character}"


def is_codepoint_surrogate(codepoint: int) -> bool:
    """
    Return if a codepoint is part of a high or low surrogate pair.

    :param codepoint: An integer of the character's codepoint.
    :return: If it's within ``0xD800..=0xDC00``.
    """
    return 0xD800 <= codepoint <= 0xDC00



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""A bitmap provider, handling PNG glyphs."""
import collections.abc
import os
import typing

import PIL.Image
import PIL.ImageChops
import unirange

import mcfonts.constants
import mcfonts.glyphs
import mcfonts.providers.base
import mcfonts.utils.image
import mcfonts.utils.schemas
import mcfonts.utils.unicode


class BitmapProvider(mcfonts.providers.base.Provider):
    """
    The ``bitmap`` provider defines simple bitmap glyphs from a texture.

    It references a single PNG file and declares a "chars" list of strings.

    The characters inside each string entry inside the charlist will be
    mapped to the corresponding positions in the referenced file.

    The texture is evenly divided according to the number of entries and the length of these entries.
    """

    def __init__(
        self,
        provider: mcfonts.utils.types.TypedProviderBitmap,
        origin: os.PathLike[str] | str,
        resource: PIL.Image.Image | None,
    ) -> None:
        self.contents = provider
        self.resource: PIL.Image.Image | None = resource
        self.glyph_cell = (0, 0)
        self.contents["height"] = self.get_contents().get("height", 8)  # Default height fallback
        if self.resource:
            if len(self.contents["chars"]) > 0:
                self.glyph_cell = (
                    self.resource.width // len(self.contents["chars"][0] or ""),
                    self.resource.height // len(self.contents["chars"]),
                )
            else:
                self.glyph_cell = (self.resource.width, self.resource.height)
        self.glyphs: dict[str, mcfonts.glyphs.BitmapGlyph] = dict(self.construct_glyphs())
        super().__init__("bitmap", origin, set(self.yield_characters()))
        self.validate()

    def pretty_print(self) -> str:
        return f"{self.get_icon()} bitmap: {self.get_contents().get('file', 'no resource')}, {len(self.glyphs):,} chars"

    def validate(self) -> None:
        validate_charlist(self.get_contents()["chars"], self.resource, self.glyph_cell)
        if self.resource:
            height = self.get_contents()["height"]
            glyph_height = self.glyph_cell[1]
            if glyph_height % height != 0:
                mcfonts.logger.warning(
                    (
                        f"Provider has an invalid height ({height}), "
                        f"it must be a multiple of the glyph's height ({glyph_height})"
                    ),
                )
            if self.glyph_cell[0] > 256 or self.glyph_cell[1] > 256:
                # The limit for a character is 256, anything beyond is an error in atlas-stitching.
                # Will still try to make it work, but make no guarantee.
                mcfonts.logger.warning(
                    (
                        "Provider has excessively large glyph size, maximum character cell size is 256, "
                        f"but got {max(self.glyph_cell[0], self.glyph_cell[1])}."
                    ),
                )

    def print_info(self, table_chars: bool = True) -> None:
        """
        Print information about this provider.

        :param table_chars:
            Whether to print a 'chars' list as a square table, or as a simple string.
        """
        super().print_info()
        print(f"File: {self.get_contents().get('file', mcfonts.constants.UNKNOWN_FIELD)}")
        print(f"Height: {self.get_contents().get('height', mcfonts.constants.UNKNOWN_FIELD)}")
        print(f"Ascent: {self.get_contents().get('ascent')}")
        if table_chars:
            print(f"Chars: ({len(self.contents['chars'][:1]):,}x{len(self.contents['chars']):,})")
            for charline in self.contents["chars"]:
                print(f"\t{' '.join(charline)}")
        else:
            print(f"Chars: {' '.join(''.join(self.contents['chars']))}")
        print(f"Count: {sum([len(x) for x in self.contents['chars']]):,}")

    def construct_glyphs(
        self,
    ) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.BitmapGlyph]]:
        if not self.resource:
            return
        for y_index, character_line in enumerate(self.contents["chars"]):
            if is_charline_empty(character_line):
                continue
            for x_index, character in enumerate(character_line):
                if character in mcfonts.constants.PADDING_CHARS:
                    continue

                height = self.contents["height"]
                yield character, mcfonts.glyphs.BitmapGlyph(
                    self.resource.crop(
                        (
                            x_index * self.glyph_cell[0],
                            y_index * self.glyph_cell[1],
                            (x_index + 1) * self.glyph_cell[0],
                            (y_index + 1) * self.glyph_cell[1],
                        )
                    ),
                    (self.contents["ascent"] - height) * (self.glyph_cell[1] // height),
                    height,
                )

    def yield_glyphs(self) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.BitmapGlyph | None]]:
        yield from self.glyphs.items()

    def yield_characters(self) -> collections.abc.Iterable[str]:
        for charline in self.contents["chars"]:
            yield from charline

    def get_glyph_for_character(self, character: str) -> mcfonts.glyphs.BitmapGlyph | None:
        """
        Extract a desired character's sheet from this provider.

        Glyph textures that are mapped to spaces or null padding return None.
        If there's more than one sheet mapped to a character (duplicate characters),
        the first occurance will be returned.

        :param character: A single character.
        :returns: A :class:`PIL.Image.Image`, or None if no glyph exists.
        """
        # Ignore non-bitmap provider, or if codepoint is null or space.
        # (glyphs can't be assigned for padding characters).
        if character in mcfonts.constants.PADDING_CHARS or self.resource is None:
            return None
        return self.glyphs.get(character)

    def reload_to_monochrome(self) -> bool:
        """
        Replace the resource used in this provider with a grayscale version.

        If the resource is already grayscale or if there's no resource, False will be returned.
        Otherwise, True.

        This modifies the resource of this provider in place, and **can't be undone**.

        :returns: Boolean of success; if there's a resource, and it isn't already grayscale.
        """
        if self.resource:
            if self.resource.mode == "LA":
                return False
            self.resource = self.resource.convert("LA")
            return True
        return False

    def get_icon(self) -> str:
        return "🖼"

    def get_pack_versions(self) -> tuple[int, None]:
        return 3, None

    def yield_glyphs_in_unirange(
        self, unirange_notation: str
    ) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.BitmapGlyph | None]]:
        for character in unirange.unirange_to_characters(unirange_notation):
            yield character, self.glyphs.get(character)

    def get_contents(self) -> mcfonts.utils.types.TypedProviderBitmap:
        return self.contents


def pad_chars_list(
    chars_list: list[str],
    pad_amount: int = 0,
    padding_character: str = " ",
    pad_from: typing.Literal["left", "right"] = "right",
) -> list[str]:
    """
    Pad `charlist` so that all its strings have the same length.

    >>> pad_chars_list(["abc", "de"])
    ['abc', 'de ']
    >>> pad_chars_list(["abc", "defg"])
    ['abc ', 'defg']

    :param chars_list: Character list to pad.
    :param pad_amount: What to pad the characters to; the ideal width.
        If this is 0, this will be determined from the largest length of strings in `charlist`.
    :param padding_character: The character to pad with, default is a space.
    :param pad_from: Direction to pad the string from, left => "___Text", right => "Text___".
    :returns: The padded charlist list.
    """
    if pad_from not in {"left", "right"}:
        raise ValueError('pad_from must be "left" or "right"')
    if pad_amount == 0:
        # Get the largest string based on length
        pad_amount = len(max(chars_list, key=len))
    for index, charline in enumerate(chars_list):
        if len(charline) != pad_amount:
            if pad_from == "left":
                # Rjust means justify to right
                chars_list[index] = charline.rjust(pad_amount, padding_character[0])
            else:
                # Ljust means justify to left
                chars_list[index] = charline.ljust(pad_amount, padding_character[0])
    return chars_list


def assert_charlist_is_equal(charlist: list[str]) -> tuple[bool, int]:
    """
    Ensure that every string inside `charlist` has the same length as the first string.

    >>> assert_charlist_is_equal(["abc","de"])
    (False, 3)
    >>> assert_charlist_is_equal(["abc","def"])
    (True, 3)
    >>> assert_charlist_is_equal([])
    (True, 0)
    >>> assert_charlist_is_equal([""])
    (True, 0)

    :param charlist: The charlist of strings to check lengths of.
    :returns: A tuple of (if the lengths are equal, the expected length).
    """
    if len(charlist) == 0:
        return True, 0
    ideal = len(max(charlist, key=len))
    for charline in charlist:
        if len(charline) != ideal:
            return False, ideal
    return True, ideal


def resource_to_charlist(resource: PIL.Image.Image, character: str, cell_bounds: tuple[int, int] = (8, 8)) -> list[str]:
    """
    Given a path to a resource and a starting character, return an appropriate charlist.

    Return a provider that has a charlist that correctly encompasses all chars covered by
    the resource. Glyphs in the resource that are empty are skipped.

    :param resource: The resource to grab textures from.
    :param character:
        A single character.
        The codepoint of this will increase by 1 on every glyph.
    :param cell_bounds:
        The dimensions of an individual glyph in `resource`.
        Glyph dimensions must be the same throughout the whole of `resource`.
    :returns: A `chars` list.
    """
    charlist: list[str] = []
    charline: list[str] = []
    codepoint = ord(character)
    height = resource.height // cell_bounds[1]
    width = resource.width // cell_bounds[0]
    for height_offset in range(height):
        for width_offset in range(width):
            if mcfonts.utils.image.is_image_empty(
                resource.crop(
                    (
                        width_offset * cell_bounds[0],
                        height_offset * cell_bounds[1],
                        (width_offset + 1) * cell_bounds[0],
                        (height_offset + 1) * cell_bounds[1],
                    )
                )
            ):
                charline.append(" ")
            else:
                charline.append(chr(codepoint))
            codepoint += 1
        charlist.append("".join(charline))
        charline.clear()
    return charlist


def validate_charlist(
    charlist: list[str],
    resource: PIL.Image.Image | None = None,
    cell_bounds: tuple[int, int] = (8, 8),
) -> None:
    r"""
    Given `charlist`, ensure that it's valid and doesn't have any possible issues.

    Issues include:

    * Empty lines,
    * Uneven lines,
    * Duplicate characters,
    * \*Padding chars in places where glyph exists in `resource`,
    * \*Characters that will have an empty glyph in `resource`.

    \* Optional, requires `resource`.

    :param charlist: A charlist to validate.
    :param resource: An optional :class:`PIL.Image.Image` to validate glyph textures against.
    :param cell_bounds: Tuple of (glyph width, glyph height) for each glyph in `resource`.
    """
    char_traces: dict[str, tuple[int, int]] = {}
    for y_index, charline in enumerate(charlist):
        if not (len_charline := len(charline)):
            mcfonts.logger.warning(f"Charline {y_index:,} is empty.")
        elif not (charlist_length := assert_charlist_is_equal(charlist))[0]:
            mcfonts.logger.warning(
                (
                    "Uneven amount of characters in charlist,"
                    f"expected {charlist_length[1]:,}, got {len_charline:,} on line {y_index:,}."
                ),
            )
        for x_index, character in enumerate(charline):
            if character not in mcfonts.constants.PADDING_CHARS and character in char_traces:
                trace = char_traces[character]
                mcfonts.logger.warning(
                    (
                        f"Duplicate character {mcfonts.utils.unicode.pretty_print_character(character)} "
                        f"on charline {y_index:,} index {x_index:,}. "
                        f"Character was already defined at charline {trace[0]:,} index {trace[1]:,}."
                    ),
                )
            else:
                char_traces[character] = (y_index, x_index)
            if resource:
                if mcfonts.utils.image.is_image_invisible(
                    resource.crop(
                        (
                            x_index * cell_bounds[0],
                            y_index * cell_bounds[1],
                            (x_index + 1) * cell_bounds[0],
                            (y_index + 1) * cell_bounds[1],
                        )
                    )
                ) and not mcfonts.utils.unicode.is_character_invisible(character):
                    mcfonts.logger.debug(
                        (
                            "Empty glyph for character "
                            f"{mcfonts.utils.unicode.pretty_print_character(character)} on "
                            f"charline {y_index:,} index {x_index:,}."
                        ),
                    )


def fit_chars_into_charlist(chars: list[str], charlist_length: int = 16) -> collections.abc.Iterable[str]:
    """
    Given a list of chars, fit them into a charlist whose width is equal to `charlist_length`.

    >>> list(fit_chars_into_charlist(["thisisareallylongcharlist"]))
    ['thisisareallylon', 'gcharlist']
    >>> list(fit_chars_into_charlist(["thisisareallylongcharlist"], 5))
    ['thisi', 'sarea', 'llylo', 'ngcha', 'rlist']

    :param chars: A list of chars.
    :param charlist_length: The width to make each line of characters in the charlist equal to.
    :returns: Yield of lines in a charlist.
    """
    chars = list(charlist_to_chars(chars))
    for i in range(0, len(chars), charlist_length):
        yield "".join(character for character in chars[i : i + charlist_length])


def charlist_to_chars(charlist: list[str]) -> collections.abc.Iterable[str]:
    """
    Given `charlist`, yield all the characters that charlist covers.

    >>> chars = ['agtg', 'b', '5', 'c', 'd', 'e', 'f', '0', '1', '2', '3', '4', '5', '6', '7', '8']
    >>> list(charlist_to_chars(chars))
    ['a', 'g', 't', 'g', 'b', '5', 'c', 'd', 'e', 'f', '0', '1', '2', '3', '4', '5', '6', '7', '8']

    :param charlist: The charlist.
    :returns: A yield of characters.
    """
    for charline in charlist:
        # iterating over lines, charline is a string
        for character in charline:
            # iterating over chars, char is a str:len=1
            yield character


def is_charline_empty(charline: str) -> bool:
    r"""
    Given `charline`, return if it contains only spaces or null bytes.

    >>> is_charline_empty("\0\0\0\x20\x20\x20\0\0")
    True
    >>> is_charline_empty("         ")
    True
    >>> is_charline_empty("      xxx")
    False

    :param charline: A single string, likely part_notation of a wider charlist.
    :returns: If `charline` is all spaces or null bytes.
    """
    return all(x in mcfonts.constants.PADDING_CHARS for x in charline)


def charlist_from_unicode_range(start: str, end: str, width: int = 16) -> list[str]:
    r"""
    Return a character list that contains characters from `start` to `end` in order.

    Given a starting character `start`, and an ending character `end`,
    return a charlist that contains these characters in order, and whose width is equal to `width`.

    >>> charlist_from_unicode_range("\u2600", "\u26ff")
    [
        '☀☁☂☃☄★☆☇☈☉☊☋☌☍☎☏',
        '☐☑☒☓☔☕☖☗☘☙☚☛☜☝☞☟',
        '☠☡☢☣☤☥☦☧☨☩☪☫☬☭☮☯',
        '☰☱☲☳☴☵☶☷☸☹☺☻☼☽☾☿',
        '♀♁♂♃♄♅♆♇♈♉♊♋♌♍♎♏',
        '♐♑♒♓♔♕♖♗♘♙♚♛♜♝♞♟',
        '♠♡♢♣♤♥♦♧♨♩♪♫♬♭♮♯',
        '♰♱♲♳♴♵♶♷♸♹♺♻♼♽♾♿',
        '⚀⚁⚂⚃⚄⚅⚆⚇⚈⚉⚊⚋⚌⚍⚎⚏',
        '⚐⚑⚒⚓⚔⚕⚖⚗⚘⚙⚚⚛⚜⚝⚞⚟',
        '⚠⚡⚢⚣⚤⚥⚦⚧⚨⚩⚪⚫⚬⚭⚮⚯',
        '⚰⚱⚲⚳⚴⚵⚶⚷⚸⚹⚺⚻⚼⚽⚾⚿',
        '⛀⛁⛂⛃⛄⛅⛆⛇⛈⛉⛊⛋⛌⛍⛎⛏',
        '⛐⛑⛒⛓⛔⛕⛖⛗⛘⛙⛚⛛⛜⛝⛞⛟',
        '⛠⛡⛢⛣⛤⛥⛦⛧⛨⛩⛪⛫⛬⛭⛮⛯',
        '⛰⛱⛲⛳⛴⛵⛶⛷⛸⛹⛺⛻⛼⛽⛾⛿'
    ]
    >>> charlist_from_unicode_range(" ", "\xff", 8)
    [
        ' !"#$%&'',
        '()*+,-./',
        '01234567',
        '89:;<=>?',
        '@ABCDEFG',
        'HIJKLMNO',
        'PQRSTUVW',
        'XYZ[\\]^_',
        '`abcdefg',
        'hijklmno',
        'pqrstuvw',
        'xyz{|}~\x7f',
        '\x80\x81\x82\x83\x84\x85\x86\x87',
        '\x88\x89\x8a\x8b\x8c\x8d\x8e\x8f',
        '\x90\x91\x92\x93\x94\x95\x96\x97',
        '\x98\x99\x9a\x9b\x9c\x9d\x9e\x9f',
        '\xa0¡¢£¤¥¦§',
        '¨©ª«¬\xad®¯',
        '°±²³´µ¶·',
        '¸¹º»¼½¾¿',
        'ÀÁÂÃÄÅÆÇ',
        'ÈÉÊËÌÍÎÏ',
        'ÐÑÒÓÔÕÖ×',
        'ØÙÚÛÜÝÞß',
        'àáâãäåæç',
        'èéêëìíîï',
        'ðñòóôõö÷',
        'øùúûüýþÿ'
    ]

    :param start: The starting single character.
    :param end:
        The ending single character.
        Codepoint must be greater than the codepoint of `start`.
    :param width: The number of characters to put in one row of the returned charlist.
    :returns: A charlist, with each string's length equal to `width`.
    """
    start_codepoint = ord(start)
    end_codepoint = ord(end)
    if end_codepoint <= start_codepoint:
        return []
    return list(fit_chars_into_charlist([chr(c) for c in range(start_codepoint, end_codepoint + 1)], width))



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""A filter is a class used to control what characters and providers are included or not included when exporting."""
from __future__ import annotations

import enum

import unirange

import mcfonts.providers.base


@enum.unique
class Policy(enum.IntEnum):
    """How to handle filtered characters or providers."""

    INCLUDE = 0
    """Include **only** the predicate matches."""
    EXCLUDE = 1
    """Include **none** of the predicate matches."""


class Filter:
    """
    A filter used in exporting; this controls which glyphs are exported and under which conditions.

    Can either act as either an excluding filter, or an inclusive filter.
    """

    def __init__(
        self,
        filtered_characters: set[str] | str | None = None,
        filtered_providers: set[type[mcfonts.providers.base.Provider]] | None = None,
        character_policy: Policy = Policy.EXCLUDE,
        provider_policy: Policy = Policy.EXCLUDE,
    ):
        """
        Construct a filter.

        :param filtered_characters: An optional string or set of strings of the character to add to the filter.
        :param filtered_providers: An optional list of the types of providers to add to the filter.
        :param character_policy: The policy to use with regards to the character filter. Can also be an integer.
        :param provider_policy: The policy to use with regards to the provider filter. Can also be an integer.
        """
        if filtered_characters is None:
            filtered_characters = set()
        if isinstance(filtered_characters, str):
            filtered_characters = unirange.unirange_to_characters(filtered_characters)
        if filtered_providers is None:
            filtered_providers = set()

        assert filtered_characters is not None and not isinstance(filtered_characters, str)  # Mypy
        self.filtered_chars = filtered_characters
        self.char_policy: Policy = character_policy
        self.filtered_providers: set[type[mcfonts.providers.base.Provider]] = filtered_providers
        self.provider_policy: Policy = provider_policy

    def filter_character(self, character: str) -> bool:
        """
        Add `character` to the character filter.

        :param character: A string of 1 character.
        :returns: Boolean of success; True if character was added, False if it was already present.
        """
        if character in self.filtered_chars:
            return False
        self.filtered_chars.add(character)
        return True

    def unfilter_character(self, character: str) -> bool:
        """
        Remove `character` from the character filter.

        :param character: A string of 1 character.
        :returns: Boolean of success; True if character was removed, False if it wasn't present.
        """
        if character not in self.filtered_chars:
            return False
        self.filtered_chars.remove(character)
        return True

    def filter_provider(self, provider: type[mcfonts.providers.base.Provider]) -> bool:
        """
        Add `provider` to the provider filter.

        :param provider: A type of :data:`~mcfonts.providers.base.Provider`.
        :returns: Boolean of success; True if the provider was added, False if it was already present.
        """
        if provider in self.filtered_providers:
            return False
        self.filtered_providers.add(provider)
        return True

    def unfilter_provider(self, provider: type[mcfonts.providers.base.Provider]) -> bool:
        """
        Remove `provider` from the provider filter.

        :param provider: A type of :data:`~mcfonts.providers.base.Provider`.
        :returns: Boolean of success; True if provider was removed, False if it was never present.
        """
        if provider not in self.filtered_providers:
            return False
        self.filtered_providers.remove(provider)
        return True

    def set_character_policy(self, new_char_policy: Policy) -> None:
        """
        Set the character policy for how to handle filtering characters.

        :param new_char_policy: A value from the :class:`Policy` enum.
        """
        self.char_policy = new_char_policy

    def set_provider_policy(self, new_provider_policy: Policy) -> None:
        """
        Set the provider policy for how to handle filtering providers.

        :param new_provider_policy: A value from the :class:`Policy` enum.
        """
        self.provider_policy = new_provider_policy

    def check_character(self, character: str) -> bool:
        """
        Determine how `character` should be handled.

        - If character is **in** filtered characters **and** character policy is EXCLUDE, or
        - If character is **not in** filtered characters **and** character policy is INCLUDE:
            - *Skip character*

        :param character: A string of a character to determine filtering for.
        :returns: A boolean; True to skip, False to include normally.
        """

        return bool(
            self.filtered_chars
            and (
                (self.char_policy == Policy.EXCLUDE and character in self.filtered_chars)
                or (self.char_policy == Policy.INCLUDE and character not in self.filtered_chars)
            )
        )

    def check_provider(self, provider: type[mcfonts.providers.base.Provider]) -> bool:
        """
        Determine how `provider` should be handled.

        - If provider is **in** filtered providers **and** provider policy is EXCLUDE, or
        - If provider is **not in** filtered providers **and** provider policy is INCLUDE:
            - *Skip provider*

        :param provider: The type of the provider to determine filtering for; this is not an instance.
        :returns: A boolean; True to skip, False to include normally.
        """
        return bool(
            self.filtered_providers
            and (
                (provider in self.filtered_providers and self.provider_policy == Policy.EXCLUDE)
                or (provider not in self.filtered_providers and self.provider_policy == Policy.INCLUDE)
            )
        )



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""A reference provider, including other fonts."""
import collections.abc
import os
import typing

import jsonschema
import unirange

import mcfonts.exceptions
import mcfonts.glyphs
import mcfonts.providers.base
import mcfonts.providers.bitmap
import mcfonts.providers.legacy_unicode
import mcfonts.providers.space
import mcfonts.providers.ttf
import mcfonts.providers.unihex
import mcfonts.utils.schemas

AnyProviderNotReference = (
    mcfonts.providers.bitmap.BitmapProvider
    | mcfonts.providers.space.SpaceProvider
    | mcfonts.providers.legacy_unicode.LegacyUnicodeProvider
    | mcfonts.providers.ttf.TTFProvider
    | mcfonts.providers.unihex.UnihexProvider
)
"""
A UnionType that does not include :class:`ReferenceProvider`.
"""

PROVIDER_SHORTCUT_MAP: dict[typing.Type[AnyProviderNotReference], str] = {
    mcfonts.providers.bitmap.BitmapProvider: "bt",
    mcfonts.providers.space.SpaceProvider: "sp",
    mcfonts.providers.legacy_unicode.LegacyUnicodeProvider: "lu",
    mcfonts.providers.ttf.TTFProvider: "ttf",
    mcfonts.providers.unihex.UnihexProvider: "uh",
}


class ReferenceProvider(mcfonts.providers.base.Provider):
    """
    The ``reference`` provider is used to include and load another font only once.

    It has one field, ``id``, which points to a font JSON to include.
    If more than 1 reference provider points to the same font JSON, it will be ignored.
    """

    def __init__(
        self,
        provider: mcfonts.utils.types.TypedProviderReference,
        origin: os.PathLike[str] | str,
        children: list[AnyProviderNotReference],
    ) -> None:
        if isinstance(children, self.__class__):
            raise mcfonts.exceptions.ReferenceChildError("a child provider cannot be another reference provider")
        self.contents = provider
        self.children = children

        super().__init__("reference", origin, set(self.yield_characters()))
        self.validate()

    def validate(self) -> None:
        jsonschema.validate(self.contents, mcfonts.utils.schemas.SCHEMA_PROVIDER_REFERENCE)

    def print_info(self) -> None:
        super().print_info()
        children = list(self.yield_children())
        print(f"Children: ({len(children)})")
        for child in children:
            print(f"\t- {str(child.__class__)}, {len(child.chars_covered):,} chars")

    def construct_glyphs(self) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.Glyph]]:
        for provider in self.yield_children():
            yield from provider.construct_glyphs()

    def yield_children(self) -> collections.abc.Iterable[AnyProviderNotReference]:
        """
        Yield all the providers that this reference provider is referring to.

        :returns: A yield of the contained providers.
        """
        yield from self.children

    def pretty_print(self) -> str:
        children_formatted = [
            PROVIDER_SHORTCUT_MAP.get(child.__class__, mcfonts.constants.UNKNOWN_FIELD) for child in self.children
        ]
        return f"{self.get_icon()} reference: {len(self.children):,} children: [{', '.join(children_formatted)}]"

    def yield_glyphs(self) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.Glyph | None]]:
        for child in self.yield_children():
            yield from child.yield_glyphs()

    def yield_characters(self) -> collections.abc.Iterable[str]:
        for child in self.children:
            yield from child.yield_characters()

    def get_icon(self) -> str:
        return "↩"

    def get_pack_versions(self) -> tuple[int, None]:
        return 15, None

    def yield_glyphs_in_unirange(
        self, unirange_notation: str
    ) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.Glyph | None]]:
        for child in self.yield_children():
            for character, glyph in zip(unirange.unirange_to_characters(unirange_notation), child.yield_glyphs()):
                assert isinstance(character, str) and isinstance(glyph, mcfonts.glyphs.Glyph)
                yield character, glyph

    def get_contents(self) -> mcfonts.utils.types.TypedProviderReference:
        return self.contents



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""
Various glyph formats that can be present in a font.

Each vary in their storage formats and methods of manipulating contained data.
"""
from __future__ import annotations

import abc
import dataclasses

import fontTools.pens.t2CharStringPen
import PIL.Image

import mcfonts.constants
import mcfonts.exceptions
import mcfonts.utils.image
import mcfonts.utils.unihex

BITMAP_EXPORT_PIXEL_THRESHOLD = 180
UNIHEX_EXPORT_PIXEL_ON = "1"
SPACE_EXPORT_MAX_ADVANCE = 12500


class Glyph(abc.ABC):
    """The base Glyph class. Should never be called or instantiated."""

    @abc.abstractmethod
    def get_width(self) -> int:
        """Width of the glyph."""

    @abc.abstractmethod
    def get_ascent(self) -> int:
        """Ascent of the glyph."""

    @abc.abstractmethod
    def get_height(self) -> int:
        """Height of the glyph. This is not always equal to the glyph's image, if applicable."""

    @abc.abstractmethod
    def get_program(self) -> list[str | int] | None:
        """A Type 2 charstring program. This is used in exporting and can be cached for better performance."""

    @abc.abstractmethod
    def construct_program(self, pixel_size: int) -> list[str | int] | None:
        """
        Create a Type 2 charstring program.

        If the glyph is empty or has no white pixels, None will be returned.
        Otherwise, a list *(the program)* will be returned.

        .. warning::

            These charstrings are **not** optimized or checked for overlaps.
            In FontForge, use :menuselection:`E&lement -> O&verlap -> &Remove Overlap` manually afterward.

        :param pixel_size:
            How big each pixel should be.
        :returns:
            A list of strings of the glyph's program, or None if there was no pixel data in ``glyph``.
        """


@dataclasses.dataclass
class BitmapGlyph(Glyph):
    """
    A bitmap glyph.

    Must contain an image and ascent.
    Height, ascent, and bearings are optional.
    Bearings and width are lazily-loaded.
    """

    image: PIL.Image.Image
    """Required image."""
    ascent: int
    """Required ascent."""
    height: int | None = None
    """Optional height. Lazy default is 8 if None."""
    width: int | None = None
    """Optional width. Lazy default is `self.image.width`."""
    bearings: tuple[int, int] | None = None
    """Optional padding from the edges of the canvas."""
    program: list[str | int] | None = None
    """Optional Type 2 charstring program of the glyph."""

    def get_bearings(self) -> tuple[int, int]:
        """
        From :func:`mcfonts.utils.resources.get_image_bearings`:

        | Bearings are a "padding" from the edge of the canvas to image pixels.
        | Left bearing is the distance from the left edge of the canvas to the most-left pixel data.
        | Right bearing is the distance from the right edge of the canvas to the most-right pixel data.

        This is lazily-loaded.
        """
        if self.bearings is None:
            self.bearings = mcfonts.utils.image.get_image_bearings(self.image)
        return self.bearings

    def get_width(self) -> int:
        """
        Width of the glyph's image.

        This is lazily-loaded.
        """
        if self.width is None:
            self.width = self.image.width
        return self.width

    def get_image(self) -> PIL.Image.Image:
        """The image of the glyph."""
        if self.image.mode not in {"RGBA", "LA"}:
            self.image = self.image.convert("RGBA")
        return self.image

    def get_height(self) -> int:
        """
        The height of the glyph; this is not necessarily the glyph's image width.

        This is lazily-loaded.
        """
        if self.height is None:
            self.height = 8
        return self.height

    def get_ascent(self) -> int:
        """The glyph's ascent."""
        return self.ascent

    def get_program(self) -> list[str | int] | None:
        if self.program is None:
            self.program = self.construct_program(
                # Em size (1000) divided by 8 (standard width) = 125,
                # divide this by scale (glyph height / JSON height) to get how big each "pixel" will translate to.
                125
                // self.image.height
                // self.get_height()
            )
        return self.program

    def construct_program(self, pixel_size: int) -> list[str | int] | None:
        if (bearings := self.get_bearings()) == (0, 0):
            return None

        pen = fontTools.pens.t2CharStringPen.T2CharStringPen(0, {})
        ascent = self.get_ascent()
        height = self.get_image().height
        glyph_width = self.get_image().width

        for index, pixel in enumerate(self.get_image().convert("LA").getdata(1)):
            if pixel >= BITMAP_EXPORT_PIXEL_THRESHOLD:
                x_real = index % glyph_width
                y_real = height - (index // glyph_width) + ascent
                pen.moveTo(
                    (
                        x_real * pixel_size,
                        y_real * pixel_size,
                    )
                )  # Left X, top right.

                pen.lineTo(
                    (
                        (x_real + 1) * pixel_size,
                        y_real * pixel_size,
                    )
                )  # Down Y, bottom left.

                pen.lineTo(
                    (
                        (x_real + 1) * pixel_size,
                        (y_real - 1) * pixel_size,
                    )
                )  # Right X, bottom right.

                pen.lineTo(
                    (
                        x_real * pixel_size,
                        (y_real - 1) * pixel_size,
                    )
                )  # Done with pixel.

                pen.closePath()
        program: list[int | str] = pen.getCharString().program
        program[0] = (bearings[1] + 1) * pixel_size  # Set the width correctly.
        return program


@dataclasses.dataclass
class SpaceGlyph(Glyph):
    """
    A space glyph.

    Contans one required field, the glyph's width.
    """

    width: int
    """The width of the glyph."""
    program: list[str | int] | None = None
    """Optional Type 2 charstring program of the glyph."""

    def get_width(self) -> int:
        return self.width

    def get_ascent(self) -> int:
        return 1

    def get_height(self) -> int:
        return 8

    def get_program(self) -> list[str | int] | None:
        if self.program is None:
            self.program = self.construct_program(125)
        return self.program

    def construct_program(self, pixel_size: float) -> list[str | int] | None:
        return [int(min(abs((self.get_width() * pixel_size)), SPACE_EXPORT_MAX_ADVANCE)), "endchar"]


@dataclasses.dataclass
class UnihexGlyph(Glyph):
    """
    A unihex glyph.

    Must contain a bit string.
    Width, image, and bearings are optional.
    Width, image, and bearings are lazily-loaded.

    Unihex glyphs are always 16 pixels tall.
    """

    bit_string: str
    """The bit string; does not include the `codepoint:` starting portion."""
    width: int | None
    """The width of the glyph. Equal to ``len(self.bit_string) // 4``."""
    image: PIL.Image.Image | None
    """
    An optional image of the glyph, comparable to a :class:`BitmapGlyph`.
    If possible, try not to use this.
    """
    bearings: tuple[int, int] | None
    """An optional tuple of the glyph's bearings, comparable to a :class:`BitmapGlyph`."""
    size_override: tuple[int, int] | None
    """
    An optional tuple of the glyph's size override.
    
    The first value is the starting column, and the second value is the ending column.
    When exporting, data outside of these columns will be discarded.
    If this is None, this is lazily-calculated to be ``(0, self.get_width())``.
    """
    program: list[str | int] | None = None
    """Optional Type 2 charstring program of the glyph."""

    def get_width(self) -> int:
        if self.width is None:
            self.width = len(self.bit_string) >> 2
        return self.width

    def get_image(self) -> PIL.Image.Image:
        """
        The image of the glyph.

        This is lazily-loaded.
        """
        if self.image is None:
            self.image = mcfonts.utils.unihex.bit_string_to_image(self.bit_string)
        return self.image

    def get_bit_string(self) -> str:
        """The bit string is the portion after the ``codepoint:``."""
        return self.bit_string

    def get_bearings(self) -> tuple[int, int]:
        """See :func:`mcfonts.utils.unihex.get_unihex_bearings`."""
        if self.bearings is None:
            self.bearings = mcfonts.utils.unihex.get_unihex_bearings(self.get_bit_string())
        return self.bearings

    def get_ascent(self) -> int:
        return 1

    def get_height(self) -> int:
        return 8

    def get_program(self) -> list[str | int] | None:
        if self.program is None:
            self.program = self.construct_program(
                # Em size (1000) divided by 10 = 100
                # divide this by scale (glyph height / JSON height) to get how big each "pixel" will translate to.
                125
            )
        return self.program

    def construct_program(self, pixel_size: int) -> list[str | int] | None:
        pen = fontTools.pens.t2CharStringPen.T2CharStringPen(0, {})
        glyph_width = self.get_width()
        size = self.get_size_override()
        index = 0
        bearings = self.get_bearings()
        for row in mcfonts.utils.unihex.bit_string_to_rows(self.get_bit_string()):
            for pixel in row:
                if pixel == UNIHEX_EXPORT_PIXEL_ON:
                    x_real = index % glyph_width
                    # - 1 to correct for baseline
                    y_real = (index // glyph_width) - 1
                    if x_real < min(size[0], 0) or x_real > min(size[1], glyph_width):
                        # Size overrides say to cut it off
                        continue
                    x_real -= bearings[0]
                    pen.moveTo(
                        (
                            x_real * pixel_size,
                            y_real * pixel_size,
                        )
                    )  # Left X, top right.

                    pen.lineTo(
                        (
                            (x_real + 1) * pixel_size,
                            y_real * pixel_size,
                        )
                    )  # Down Y, bottom left.

                    pen.lineTo(
                        (
                            (x_real + 1) * pixel_size,
                            (y_real - 1) * pixel_size,
                        )
                    )  # Right X, bottom right.

                    pen.lineTo(
                        (
                            x_real * pixel_size,
                            (y_real - 1) * pixel_size,
                        )
                    )  # Done with pixel.

                    pen.closePath()
                index += 1
        program: list[int | str] = pen.getCharString().program
        program[0] = (bearings[1] - bearings[0] + 1) * pixel_size  # Set the width correctly.
        return program

    def get_size_override(self) -> tuple[int, int]:
        """Size override is a tuple of 2 integers of the starting and ending columns for a glyph."""
        if self.size_override is None:
            self.size_override = (0, self.get_width())
        return self.size_override



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""
Functions for compacting glyphs.

Compacting is the process of taking every glyph provided
by a Minecraft font provider and storing it in as little space as possible in a bitmap image.

This means:

1. Finding the largest effective dimensions out of all the characters,
2. Expanding the dimensions of all other character textures to fit those new dimensions,
3. Creating a new texture that will fit all characters in those dimensions,
4. Sequencing them one-after-another while ignoring characters with a blank texture,
5. Creating a new charlist that corresponds with the compacted texture.

This contains functions for compacting any provider(s) into individual textures
(one for each provider), and for compacting a list of providers into a single, cohesive texture.
"""
from __future__ import annotations

import math

import PIL.Image
import PIL.ImageChops

import mcfonts.glyphs
import mcfonts.utils.image


def compact_glyphs(
    glyphs: list[mcfonts.glyphs.BitmapGlyph | mcfonts.glyphs.UnihexGlyph],
    chars_in_row: int = 0,
    cell_size: tuple[int, int] = (0, 0),
    square_cells: bool = True,
) -> tuple[PIL.Image.Image, tuple[int, int]]:
    """
    Compact a list of glyphs so that they take up a minimum amount of space in an image.

    Given a list of glyphs, return an image wherein all glyphs are compacted into a single sheet, with the number of
    glyphs in one row matching ``chars_in_row`` (excluding exceptions).

    :param glyphs:
        A list of lists of [glyphs (:class:`PIL.Image.Image`), their height, and their ascent].
        Baseline can be calculated from this: :math:`height - ascent`.
    :param chars_in_row:
        How many characters to fit inside each row of the resulting sheet.
        Only positive values will set fixed rows.
        If this is negative or 0, this will be set so that the resulting sheet is square.
        By default, this is 0 (a square).
    :param cell_size:
        What size to make each glyph cell.
        If this is (0, 0), this will be set to the largest dimensions of every glyph in `glyphs`.
    :param square_cells:
        If True, each glyph's width will equal its height.
        This is based on whichever number is largest.
        If False, each glyph's width will be unrelated to its height.
    :returns:
        A tuple containing a modified `resource` with the least amount of padding between glyphs,
        and a tuple of the size of each glyph cell.
    """
    if chars_in_row <= 0:
        chars_in_row = math.ceil(math.sqrt(len(glyphs)))

    if cell_size == (0, 0):
        max_cell = mcfonts.utils.image.get_largest_effective_image_dimensions(
            list(glyph.get_image() for glyph in glyphs)
        )
    else:
        max_cell = cell_size
    if square_cells:
        maximum = max(max_cell)
        max_cell = (maximum, maximum)

    average_baseline = round(sum(glyph.get_height() - glyph.get_ascent() for glyph in glyphs) / len(glyphs))

    destination = PIL.Image.new(
        "RGBA",
        (max_cell[0] * chars_in_row, max_cell[1] * len(glyphs) // chars_in_row),
        (0, 0, 0, 0),
    )
    for index, glyph in enumerate(glyphs):
        destination.paste(
            PIL.ImageChops.offset(glyph.get_image(), 0, average_baseline - glyph.get_height() - glyph.get_ascent()),
            (
                (index % chars_in_row) * max_cell[0],
                (index // chars_in_row) * max_cell[1],
            ),
        )
    return destination, max_cell


def compact_images(
    images: list[PIL.Image.Image | None],
    chars_in_row: int = 0,
    cell_size: tuple[int, int] = (0, 0),
    square_cells: bool = True,
    include_empty_images: bool = True,
) -> tuple[PIL.Image.Image, tuple[int, int]]:
    """Similar to :func:`compact_glyphs`, except that it works on plain :class:`~PIL.Image.Image` instead."""

    max_cell = cell_size
    if not include_empty_images:
        filtered_images = [image for image in images if image is not None]
        if cell_size == (0, 0):
            max_cell = mcfonts.utils.image.get_largest_effective_image_dimensions(filtered_images)
    else:
        filtered_images = images  # type: ignore[assignment] # It's OK if images is list[Image | None] from here out

    if chars_in_row <= 0:
        chars_in_row = math.ceil(math.sqrt(len(filtered_images)))

    if square_cells:
        maximum = max(max_cell)
        max_cell = (maximum, maximum)

    destination = PIL.Image.new(
        "RGBA",
        (max_cell[0] * chars_in_row, max_cell[1] * len(filtered_images) // chars_in_row),
        (0, 0, 0, 0),
    )
    for index, glyph in enumerate(filtered_images):
        if glyph:
            destination.paste(
                glyph,
                (
                    (index % chars_in_row) * max_cell[0],
                    (index // chars_in_row) * max_cell[1],
                ),
            )
    return destination, max_cell



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""
Utilities for dealing with .hex glyphs.

This is often used in conjunction with :class:`mcfonts.providers.unihex.UnihexProvider`.
"""
from __future__ import annotations

import collections.abc

import PIL.Image
import PIL.ImageOps


def get_codepoint(hex_glyph: str) -> str:
    """
    Return the character that a .hex glyph is associated with.

    :param hex_glyph: The entire .hex glyph.
    :returns: The character portion of a .hex glyph.
    """
    return chr(int(hex_glyph.split(":", 1)[0], 16))


def get_width(bit_string: str) -> int:
    """
    Return how wide a .hex glyph's bit string is.

    Width is equivalent to ``len(bit_string) // 4``, which can be shortened to ``len(bit_string) >> 2``.

    :param bit_string: The bit string portion of a .hex glyph.
    :returns: The width of the bit string. A power of 2 above or equal to 8.
    """
    return len(bit_string) >> 2


def get_bit_string(hex_glyph: str) -> str:
    """
    Return the "bit string" portion of a .hex glyph. This is the part after the ``:``.

    >>> get_bit_string("0041:0000000018242442427E424242420000")
    '0000000018242442427E424242420000'

    :param hex_glyph: A single .hex glyph.
    :returns: The bit string portion.
    """
    return hex_glyph.split(":", 1)[1]


def bit_string_to_bytes(bit_string: str) -> bytes:
    r"""
    Return `bit_string` but with each set of "bytes" as real bytes.

    For example, "00EA" returns "\\x00\\xEA".
    This is useful for iterating over the pixels of a bit string, or converting to an image.

    >>> bit_string_to_bytes("0000000018242442427E424242420000")
    b'\x00\x00\x00\x00\x18$$BB~BBBB\x00\x00'

    :param bit_string: The bit string.
    :returns: Bytes of the bit string, with each two-character portion translated to their byte representation.
    """
    r = []
    for i in range(0, len(bit_string), 2):
        r.append(int(bit_string[i : i + 2], 16))
    return bytes(r)


def bit_string_to_image(bit_string: str) -> PIL.Image.Image:
    """
    Given `bit_string`, return a :class:`~PIL.Image.Image` that is a 1:1 representation of the glyph.

    :param bit_string: A bit string.
    :returns: An :class:`~PIL.Image.Image` of the bit string.
    """
    return PIL.Image.frombuffer("1", (get_width(bit_string), 16), bit_string_to_bytes(bit_string))


def bit_string_to_rows(bit_string: str, reverse: bool = True) -> collections.abc.Iterable[str]:
    """
    Convert `bit_string` into a yield of strings, with each character in the string being either "1" or "0".

    If `reverse` (the default), the rows will be flipped vertically.

    >>> list(bit_string_to_rows("0000000018242442427E424242420000"))
    [
        '00000000',
        '00000000',
        '01000010',
        '01000010',
        '01000010',
        '01000010',
        '01111110',
        '01000010',
        '01000010',
        '00100100',
        '00100100',
        '00011000',
        '00000000',
        '00000000',
        '00000000',
        '00000000'
    ]

    :param bit_string: The bit string.
    :param reverse:
        Whether to flip the image along the Y-axis.
        This is to match the normal behavior of iterating over the pixels of a :class:`PIL.Image.Image`,
        from the bottom-left to the top-right, row first.
    :return:
    """
    row_width_hex = len(bit_string) >> 4  # len // 16
    row_width_bin = len(bit_string) >> 2  # len // 4
    if reverse:
        for i in reversed(range(0, len(bit_string), row_width_hex)):
            yield bin(int(bit_string[i : i + row_width_hex], 16))[2:].zfill(row_width_bin)
    else:
        for i in range(0, len(bit_string), row_width_hex):
            yield bin(int(bit_string[i : i + row_width_hex], 16))[2:].zfill(row_width_bin)


def get_unihex_bearings(bit_string: str) -> tuple[int, int]:
    """
    Return the two integers of the bearings of an .hex glyph's bit string.

    Bearings are a "padding" from the edge of the canvas to image pixels.
    Left bearing is the distance from the left edge of the canvas to the most-left pixel data.
    Right bearing is the distance from the right edge of the canvas to the most-right pixel data.

    If return is (0, 0), there's no pixel data, the glyph is all spaces.

    This function is similar to :func:`mcfonts.utils.image.get_image_bearings`.

    >>> get_unihex_bearings("0000000018242442427E424242420000")
    (1, 7)

    :param bit_string: The bit string of a .hex glyph.
    :returns:
        Left bearing and right bearing.
        Returns (0, 0) if there's no pixel data.
    """
    glyph_width = get_width(bit_string)
    grid = [list(row) for row in bit_string_to_rows(bit_string)]

    left_bearing = 0
    right_bearing = glyph_width

    for x in range(glyph_width):
        if all(grid[y][x] == "0" for y in range(16)):
            left_bearing += 1
        else:
            break

    if left_bearing == glyph_width:
        # The left bearing is the whole glyph, so there's no data here anyway.
        return 0, 0

    for x_reversed in reversed(range(glyph_width)):
        if all(grid[y][x_reversed] == "0" for y in range(16)):
            right_bearing -= 1
        else:
            break

    return left_bearing, right_bearing



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""Simple utilities for working with :class:`PIL.Image.Image`."""
from __future__ import annotations

import collections.abc

import PIL.Image
import PIL.ImageOps


def is_image_empty(image: PIL.Image.Image) -> bool:
    """
    Determine if `image` has any pixel data.

    :param image: A :class:`PIL.Image.Image`.
    :returns: If `image` has pixel data.
    """
    for band in image.getextrema():  # type: ignore[no-untyped-call]
        if (isinstance(band, int) and band != 0) or (len(band) >= 2 and (band[0] != 0 or band[1] != 0)):
            return False
    return True


def is_image_invisible(image: PIL.Image.Image) -> bool:
    """
    Determine if `image` has all invisible pixels; if alpha is 0.

    :param image: A :class:`PIL.Image.Image`.
    :returns: If `image` doesn't have any full-alpha pixels.
    """
    if image.mode == "RGBA":
        return all(pixel == 0 for pixel in image.getdata(3))
    if image.mode == "LA":
        return all(pixel == 0 for pixel in image.getdata(1))
    return False


def get_largest_effective_image_dimensions(
    images: list[PIL.Image.Image], isolate_dimensions: bool = True
) -> tuple[int, int]:
    """
    Given a list of images, return the width and height of the largest images in that list.

    This function operates based on the pixels inside an image;
    it crops out empty space and then determines the width and height.
    For a function that operates on an image's true width and height, see func:`get_largest_image_dimensions`.

    :param images: An iterable of the images to iterate over.
    :param isolate_dimensions:
        If True, the maximum height and width will be calculated.
        This means that dimensions (4, 6) (12, 8) (3, 9) will be (12, 9).

        If False, the maximum height and width is dependent on the maximum width found.
        This means that dimensions (4, 6) (12, 8) (3, 9) will be (12, 8).
        It is recommended that you leave this to True.
    :returns: A tuple of (maximum width, maximum height).
    """
    max_width = 0
    max_height = 0
    for glyph in images:
        if glyph:
            if bbox := glyph.getbbox():
                if (width := bbox[2] - bbox[0]) > max_width:
                    max_width = width
                    if not isolate_dimensions:
                        max_height = width
                    continue
                if (height := bbox[3] - bbox[1]) > max_height:
                    max_height = height
    return max_width, max_height


def get_largest_image_dimensions(
    images: collections.abc.Iterable[PIL.Image.Image | None], isolate_dimensions: bool = True
) -> tuple[int, int]:
    """
    Given a list of images, return the width and height of the largest image's canvas in that list.

    This function is different from func:`get_largest_effective_image_dimensions` in that it
    does not calculate the real width and height that pixels in an image extend to. It only
    relies on the width and height of the image's *canvas* (the total space).

    :param images: A list of the glyphs to iterate over.
    :param isolate_dimensions:
        If True, the maximum height and width will be calculated separately.
        This means that dimensions (4, 6) (12, 8) (3, 9) will be (12, 9).

        If False, the maximum height and width is dependent on the maximum width found.
        This means that dimensions (4, 6) (12, 8) (3, 9) will be (12, 8).

        It is recommended that you leave this to True.
    :returns: A tuple of (maximum width, maximum height).
    """
    max_width = 0
    max_height = 0
    for glyph in images:
        if glyph and glyph.getbbox():
            if (width := glyph.width) > max_width:
                max_width = width
                if not isolate_dimensions:
                    max_height = width
                continue
            if (height := glyph.height) > max_height:
                max_height = height
    return max_width, max_height


def get_image_bearings(image: PIL.Image.Image) -> tuple[int, int]:
    """
    Return the two integers of the bearings of an image.

    Bearings are a "padding" from the edge of the canvas to image pixels.
    Left bearing is the distance from the left edge of the canvas to the most-left pixel data.
    Right bearing is the distance from the right edge of the canvas to the most-right pixel data.

    If return is (0, 0), there's no pixel data, the glyph is all spaces.

    :param image: A :class:`PIL.Image.Image` instance.
    :returns:
        Left bearing and right bearing.
        Returns (0, 0) if there's no pixel data.
    """
    try:
        bbox = image.getbbox()
        if bbox is None:
            raise TypeError
        return bbox[0], bbox[2]
    except TypeError:
        # Pure space, has no width
        return 0, 0



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""
ANSI color escapes.
The global variable ``USE_COLORS`` controls whether these will be used in outputs.

* Set :const:`USE_COLORS` = False to **disable** colored output.
* Set :const:`USE_COLORS` = True to **enable** colored output.
"""
from __future__ import annotations

USE_COLORS = True
"""
Whether to include ANSI color escapes in some prints.
Set this to False if your console/terminal does not support ANSI color escapes.
"""

_CSI = "\x1b["
BLACK_FORE = f"{_CSI}30m"
RED_FORE = f"{_CSI}31m"
GREEN_FORE = f"{_CSI}32m"
YELLOW_FORE = f"{_CSI}33m"
BLUE_FORE = f"{_CSI}34m"
MAGENTA_FORE = f"{_CSI}35m"
CYAN_FORE = f"{_CSI}36m"
WHITE_FORE = f"{_CSI}37m"
RESET_FORE = f"{_CSI}39m"

BLACK_BACK = f"{_CSI}40m"
RED_BACK = f"{_CSI}41m"
GREEN_BACK = f"{_CSI}42m"
YELLOW_BACK = f"{_CSI}43m"
BLUE_BACK = f"{_CSI}44m"
MAGENTA_BACK = f"{_CSI}45m"
CYAN_BACK = f"{_CSI}46m"
WHITE_BACK = f"{_CSI}47m"
RESET_BACK = f"{_CSI}49m"

RESET_ALL = f"{_CSI}0m"
BRIGHT = f"{_CSI}1m"
DIM = f"{_CSI}2m"


def color_number(number: float | int) -> str:
    r"""
    Given `number`, return a colorized and pretty-print version of that number.

    .. note::
        If :data:`mcfonts.colors.USE_COLORS` is False, color won't be applied.

    If `number` is negative, it will be in red.
    If `number` is positive, it will be in green.
    If `number` is zero, it will have no colors.

    >>> color_number(4)
    '\x1b[32m+4\x1b[39m'
    >>> color_number(-4)
    '\x1b[31m-4\x1b[39m'
    >>> color_number(-39924)
    '\x1b[31m-39,924\x1b[39m'

    :param number: The number, positive or negative.
    :returns: A string representing `number` with ANSI color codes.
    """
    if number < 0 and USE_COLORS:
        return f"{RED_FORE}{number:,}{RESET_FORE}"
    if number == 0 or not USE_COLORS:
        return f"{number:,}"
    if USE_COLORS:
        return f"{GREEN_FORE}+{number:,}{RESET_FORE}"
    return f"+{number:,}"



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""Command-line tool front-end for mcfonts."""
from __future__ import annotations

import argparse
import inspect
import os.path
import sys

import PIL.Image

import mcfonts.utils.resources

if __name__ == "__main__":

    def __check_dir(path: str) -> str:
        if os.path.exists(path):
            return path
        raise FileNotFoundError(path)

    INPUT_HELP = (
        "Path to folder, JSON, or ZIP file of font. "
        "If folder, this folder must be a resource pack. "
        "If JSON, this file must be a font JSON file. "
        "If ZIP, this file must be a resource pack archive. "
        "If -, stdin is used."
    )

    parser = argparse.ArgumentParser(
        description="""
        mcfonts is a versatile, fast, and extensible package for working with Minecraft fonts.
        
        The CLI front-end does not expose every possible option.
        For more in-depth usage, import the mcfonts module in Python.
        """
    )

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="mcfonts " + ".".join(str(x) for x in mcfonts.__version__),
    )

    subparsers = parser.add_subparsers(title="Functions")

    export_parser = subparsers.add_parser("export", help="export to another font format")
    export_parser.add_argument(
        "input",
        type=__check_dir,
        default=(None if sys.stdin.isatty() else sys.stdin),
        help=INPUT_HELP,
    )
    export_parser.add_argument(
        "outfile", type=argparse.FileType("wb"), default=sys.stdout, help="Where to write the file to."
    )
    export_parser.add_argument("-n", "--name", help="Name of the font.", default=mcfonts.constants.UNKNOWN_FIELD)
    export_parser.add_argument("-c", "--credits", type=bool, help="Whether to embed the mcfonts credits.", default=True)
    export_parser.add_argument(
        "-a",
        "--author",
        help="Author of the font.",
    )
    export_parser.add_argument("-l", "--license", type=bool, help="License of the font.", required=False)
    export_parser.add_argument(
        "format",
        help="What format to export to.",
        default="opentype",
        # Dynamically get all functions in mcfonts.export_formats
        choices={x[0] for x in inspect.getmembers(mcfonts.export_formats, inspect.isfunction)},
        nargs="?",
    )
    export_parser.set_defaults(export=True)

    info_parser = subparsers.add_parser("info", help="print information about a font file")
    info_parser.add_argument(
        "input",
        type=argparse.FileType(),
        default=(None if sys.stdin.isatty() else sys.stdin),
        help=INPUT_HELP,
    )
    info_parser.add_argument(
        "-t",
        "--table-chars",
        action="store_true",
        default=False,
        help="to put characters into a formatter table or put in single list",
    )
    info_parser.add_argument(
        "-s",
        "--summary-only",
        action="store_true",
        default=False,
        help="to print only a small summary",
    )
    info_parser.set_defaults(info=True)

    compact_parser = subparsers.add_parser("compact", help="fit the glyphs in the minimum amount of space needed")
    compact_parser.add_argument(
        "input",
        type=argparse.FileType("rb"),
        default=(None if sys.stdin.isatty() else sys.stdin),
        help="Path to PNG file.",
    )
    compact_parser.add_argument(
        "char-counts",
        nargs=2,
        type=int,
        default=(16, 16),
        help="Two numbers of the characters in each row and column of the input file.",
    )
    compact_parser.add_argument(
        "-c",
        "--chars-in-row",
        type=int,
        default=16,
        help="The desired number of glyphs to put in one row, set to 0 for square. Default is 16.",
    )
    compact_parser.add_argument("outfile", type=argparse.FileType("wb"), default=sys.stdout)
    compact_parser.set_defaults(compact=True)

    compare_parser = subparsers.add_parser("compare", help="compare two fonts")
    compare_parser.add_argument(
        "input1",
        type=argparse.FileType("rb"),
        default=(None if sys.stdin.isatty() else sys.stdin),
        help=INPUT_HELP,
    )
    compare_parser.add_argument(
        "input2",
        type=argparse.FileType("rb"),
        default=(None if sys.stdin.isatty() else sys.stdin),
        help=INPUT_HELP,
    )
    compare_parser.add_argument(
        "-s",
        "--swap",
        action="store_true",
        default=(None if sys.stdin.isatty() else sys.stdin),
        help="swap the order of fonts compared",
    )
    compare_parser.set_defaults(compare=True)

    vargs = vars(parser.parse_args(args=None if sys.argv[1:] else ["--help"]))

    if vargs.get("export"):
        to_write = mcfonts.from_java_ambiguous(vargs["input"]).export(
            getattr(mcfonts.export_formats, vargs["format"]) or mcfonts.export_formats.opentype,
            {
                "name": vargs["name"],
                "include_credits": vargs["include_credits"],
                "author": vargs["author"],
                "license": vargs["license"],
            },
        )
        vargs["outfile"].write(to_write)
    elif vargs.get("info"):
        mcfonts.from_java_ambiguous(vargs["input"].name).print_info(
            vargs.get("table_chars", False), vargs.get("summary_only", False)
        )
    elif vargs.get("compact"):
        resource = PIL.Image.open(vargs["input"])
        cell_size = tuple(vargs["char_counts"])
        compacted = mcfonts.compacting.compact_images(
            list(mcfonts.utils.resources.divide_resource_by_grid(resource, cell_size)),
            vargs["chars_in_row"](resource.width // cell_size[0], resource.height // cell_size[1]),
        )
        compacted[0].save(vargs["outfile"])
        print(compacted[1])
    elif vargs.get("compare"):
        font1 = mcfonts.from_java_ambiguous(vargs["input1"].name)
        font2 = mcfonts.from_java_ambiguous(vargs["input2"].name)
        if vargs.get("swap", False):
            font2.compare(font1)
        else:
            font1.compare(font2)



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""Utilities for importing provider dictionaries."""
from __future__ import annotations

import collections.abc
import copy
import json
import pathlib
import typing

import jsonschema
import PIL

import mcfonts.exceptions
import mcfonts.providers
import mcfonts.providers.reference
import mcfonts.utils


def load_all(
    providers: list[dict[str, typing.Any]],
    json_path: pathlib.Path,
    strict: bool = False,
    mode: str = "LA",
) -> collections.abc.Iterable[mcfonts.providers.base.Provider]:
    """
    Load every provider in the list `providers`.

    :param providers:
        A list of dictionaries of providers.
        This should be the value of the "providers" key in the JSON file.
    :param json_path: The path to the JSON file at which this provider JSON is from.
    :param strict: Whether to raise or warn (and continue) on exceptions.
    :param mode:
        What "mode" to load bitmap resources in.

        * ``"LA"`` - Grayscale
        * ``"RGBA"`` - Color

    :returns: A yield of Providers.
    """
    for provider in copy.deepcopy(providers):
        yield load_provider(provider, json_path, strict, mode)


def load_provider(
    provider: dict[str, typing.Any],
    json_path: pathlib.Path,
    strict: bool = False,
    mode: str = "LA",
) -> mcfonts.providers.base.Provider:
    """
    Load a specific provider dictionary and return the equivalent Provider class.
    Performs JSON schema validation.

    :param provider: A dictionary of a provider.
    :param json_path: The real path to the JSON file at which this provider JSON is from.
    :param strict: Whether to raise or warn (and continue) on exceptions.
    :param mode:
        What "mode" to load bitmap resources in.

        * ``"LA"`` - Grayscale
        * ``"RGBA"`` - Color

    :returns: A Provider instance.
    """
    provider_type = provider.get("type", "")
    try:
        if provider_type == "bitmap":
            jsonschema.validate(provider, mcfonts.utils.schemas.SCHEMA_PROVIDER_BITMAP)
            typed_bitmap_provider: mcfonts.utils.types.TypedProviderBitmap = {
                "type": "bitmap",
                "file": provider["file"],
                "height": provider.get("height", 8),
                "ascent": provider["ascent"],
                "chars": provider["chars"],
            }
            return mcfonts.providers.bitmap.BitmapProvider(
                typed_bitmap_provider,
                json_path,
                mcfonts.utils.resources.load_bitmap(
                    typed_bitmap_provider["file"], mcfonts.utils.resources.traverse_to_assets(json_path)
                ),
            )
        elif provider_type == "legacy_unicode":
            jsonschema.validate(provider, mcfonts.utils.schemas.SCHEMA_PROVIDER_LEGACY_UNICODE)
            typed_legacy_unicode_provider: mcfonts.utils.types.TypedProviderLegacyUnicode = {
                "type": "legacy_unicode",
                "template": provider["template"],
                "sizes": provider["sizes"],
            }
            return mcfonts.providers.legacy_unicode.LegacyUnicodeProvider(
                typed_legacy_unicode_provider,
                json_path,
                dict(
                    mcfonts.utils.resources.load_resources_legacy_unicode(
                        typed_legacy_unicode_provider["template"],
                        typed_legacy_unicode_provider["sizes"],
                        mcfonts.utils.resources.traverse_to_assets(json_path),
                    )
                ),
            )
        elif provider_type == "space":
            jsonschema.validate(provider, mcfonts.utils.schemas.SCHEMA_PROVIDER_SPACE)
            typed_space_provider: mcfonts.utils.types.TypedProviderSpace = {
                "type": "space",
                "advances": provider["advances"],
            }
            return mcfonts.providers.space.SpaceProvider(typed_space_provider, json_path)
        elif provider_type == "ttf":
            jsonschema.validate(provider, mcfonts.utils.schemas.SCHEMA_PROVIDER_TTF)
            typed_ttf_provider: mcfonts.utils.types.TypedProviderTTF = {
                "type": "ttf",
                "file": provider["file"],
                "shift": provider["shift"],
                "size": provider["size"],
                "oversample": provider["oversample"],
                "skip": provider["skip"],
            }
            with open(
                mcfonts.utils.resources.resolve_path(
                    typed_ttf_provider["file"], mcfonts.utils.resources.traverse_to_assets(json_path), "font"
                ),
                "rb",
            ) as open_tempfile:
                ttf = open_tempfile.read()
            return mcfonts.providers.ttf.TTFProvider(typed_ttf_provider, json_path, ttf)
        elif provider_type == "unihex":
            jsonschema.validate(provider, mcfonts.utils.schemas.SCHEMA_PROVIDER_UNIHEX)
            typed_unihex_provider: mcfonts.utils.types.TypedProviderUnihex = {
                "type": "unihex",
                "hex_file": provider["hex_file"],
                "size_overrides": provider["size_overrides"],
            }
            return mcfonts.providers.unihex.UnihexProvider(
                typed_unihex_provider,
                json_path,
                dict(
                    mcfonts.utils.resources.load_resources_unihex(
                        typed_unihex_provider["hex_file"], mcfonts.utils.resources.traverse_to_assets(json_path)
                    )
                ),
            )
        elif provider_type == "reference":
            jsonschema.validate(provider, mcfonts.utils.schemas.SCHEMA_PROVIDER_REFERENCE)
            typed_reference_provider: mcfonts.utils.types.TypedProviderReference = {
                "type": "reference",
                "id": provider["id"],
            }
            # Create any child providers first
            with open(
                temp_recursive_path := mcfonts.utils.resources.resolve_path(
                    typed_reference_provider["id"] + ".json",
                    mcfonts.utils.resources.traverse_to_assets(json_path),
                    "font",
                ),
                encoding="utf-8",
            ) as temp_recursive_file:
                data = json.load(temp_recursive_file, strict=strict)
            children = list(
                load_all(data.get("providers", {}), temp_recursive_path, strict, mode)
            )  # This acts recursively.
            if any(isinstance(child, mcfonts.providers.reference.ReferenceProvider) for child in children):
                raise mcfonts.exceptions.ReferenceChildError()
            return mcfonts.providers.reference.ReferenceProvider(
                typed_reference_provider,
                json_path,
                children,  # type: ignore[arg-type] # children will never be list[Provider]
            )
        elif provider_type in {"mcfonts:options", "options"}:
            raise mcfonts.exceptions.ProviderError("Options providers are longer accepted and must be removed.")
    except PIL.UnidentifiedImageError as exception:
        if strict:
            raise exception
        mcfonts.logger.warning(
            mcfonts.providers.base.format_provider_message(provider, "has invalid file; must be PNG.")
        )

    except FileNotFoundError as exception:
        if strict:
            # Missing files not allowed, raising exception
            raise exception
            # Missing files are ignored
        mcfonts.logger.warning(
            mcfonts.providers.base.format_provider_message(
                provider, f'file "{exception.filename}" does not exist, skipping.'
            )
        )

    raise mcfonts.exceptions.ProviderError(f'Unknown provider "{provider_type}".')



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""A legacy unicode provider, handling bitmap glyphs in a 16px x 16px grid."""
import collections.abc
import os
import re
import unicodedata

import PIL.Image
import unirange

import mcfonts.constants
import mcfonts.glyphs
import mcfonts.providers.base
import mcfonts.utils.resources
import mcfonts.utils.schemas

TEMPLATE_CHARS = {
    0: PIL.Image.frombytes("1", (3, 5), b"\xe0\xa0\xa0\xa0\xe0"),
    1: PIL.Image.frombytes("1", (3, 5), b"@\xc0@@\xe0"),
    2: PIL.Image.frombytes("1", (3, 5), b"\xe0 \xe0\x80\xe0"),
    3: PIL.Image.frombytes("1", (3, 5), b"\xe0 \xe0 \xe0"),
    4: PIL.Image.frombytes("1", (3, 5), b"\xa0\xa0\xe0  "),
    5: PIL.Image.frombytes("1", (3, 5), b"\xe0\x80\xe0 \xe0"),
    6: PIL.Image.frombytes("1", (3, 5), b"\xe0\x80\xe0\xa0\xe0"),
    7: PIL.Image.frombytes("1", (3, 5), b"\xe0    "),
    8: PIL.Image.frombytes("1", (3, 5), b"\xe0\xa0\xe0\xa0\xe0"),
    9: PIL.Image.frombytes("1", (3, 5), b"\xe0\xa0\xe0 \xe0"),
    10: PIL.Image.frombytes("1", (3, 5), b"\xe0\xa0\xe0\xa0\xa0"),
    11: PIL.Image.frombytes("1", (3, 5), b"\xc0\xa0\xc0\xa0\xc0"),
    12: PIL.Image.frombytes("1", (3, 5), b"\xe0\x80\x80\x80\xe0"),
    13: PIL.Image.frombytes("1", (3, 5), b"\xc0\xa0\xa0\xa0\xc0"),
    14: PIL.Image.frombytes("1", (3, 5), b"\xe0\x80\xe0\x80\xe0"),
    15: PIL.Image.frombytes("1", (3, 5), b"\xe0\x80\xe0\x80\x80"),
    -4: PIL.Image.frombytes(
        "1", (11, 16), b"\xff\xe0\x80 \x80 \x80 \x80 \x80 \x80 \x80 \x80 \x80 \x80 \x80 \x80 \x80 \x80 \xff\xe0"
    ),
    -6: PIL.Image.frombytes(
        "1",
        (15, 16),
        b"\xff\xfe\x80\x02\x80\x02\x80\x02\x80\x02\x80\x02\x80\x02\x80\x02\x80\x02\x80\x02\x80\x02\x80\x02\x80\x02\x80"
        b"\x02\x80\x02\xff\xfe",
    ),
}


class LegacyUnicodeProvider(mcfonts.providers.base.Provider):
    """
    The ``legacy_unicode`` provider is a "fallback" provider intended to be used as a last-resort.

    It is similar to the :class:`bitmap provider<mcfonts.providers.bitmap.BitmapProvider>`
    in that its glyphs are bitmaps.

    It uses a system of templates to create and add 16px x 16px tables of "fallback" characters.
    Each glyph is 16px x 16px pixels wide, so each page is 256px x 256px;
    16 characters on each line, 16 lines, 16 length for each glyph.

    .. warning::
        This provider is deprecated and should not be used when possible.
        Use the "unihex" provider instead.

    .. important::
        Characters above the `BMP <https://en.wikipedia.org/wiki/Plane_(Unicode)#Basic_Multilingual_Plane>`_ (U+FFFF)
        are not handled.
    """

    def __init__(
        self,
        provider: mcfonts.utils.types.TypedProviderLegacyUnicode,
        origin: os.PathLike[str] | str,
        resources: dict[str, PIL.Image.Image | bytes],
    ) -> None:
        self.contents = provider
        self.resources = resources
        self.codepages_covered: list[int] = []
        self.glyphs = dict(self.construct_glyphs())
        super().__init__(
            "legacy_unicode",
            origin,
            set(self.yield_characters()),
        )
        self.validate()

    def pretty_print(self) -> str:
        return f"{self.get_icon()} legacy_unicode: {self.get_contents().get('template', 'no template')}"

    def validate(self) -> None:
        return None  # No specific validation checks yet.

    def print_info(self) -> None:
        super().print_info()
        print(f"Template: {self.contents['template'] or mcfonts.constants.UNKNOWN_FIELD}")
        print(f"Sizes: {self.contents['sizes'] or mcfonts.constants.UNKNOWN_FIELD}")

    def construct_glyphs(self) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.BitmapGlyph]]:
        for codepage, content in self.resources.items():
            if not re.match(
                self.contents["template"].replace("\\", "\\\\").replace("(", "\\(").replace("%s", r"(\d\d)"), codepage
            ) or not isinstance(content, PIL.Image.Image):
                continue  # If template_\d\d\.png doesn't match codepage name, skip.
            if not (codepage_num := re.match(codepage, r"\d\d")):
                continue  # If template doesn't have \d\d, skip.
            for index, glyph in enumerate(mcfonts.utils.resources.divide_resource_by_grid(content, (16, 16))):
                if glyph is None:
                    continue
                yield chr((int(codepage_num.group(1), 16) << 8) + index), mcfonts.glyphs.BitmapGlyph(glyph, 1)

    def yield_glyphs(self) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.BitmapGlyph]]:
        yield from self.glyphs.items()

    def yield_characters(self) -> collections.abc.Iterable[str]:
        for key in self.glyphs.keys():
            yield str(key)

    def get_icon(self) -> str:
        return "🖳"

    def get_pack_versions(self) -> tuple[int, int]:
        return 4, 19

    def yield_glyphs_in_unirange(
        self, unirange_notation: str
    ) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.BitmapGlyph | None]]:
        for character in unirange.unirange_to_characters(unirange_notation):
            yield character, self.glyphs.get(character)

    def get_contents(self) -> mcfonts.utils.types.TypedProviderLegacyUnicode:
        return self.contents


def to_advances(glyphsizes: bytes, match_unicode_category: list[str] | None = None) -> dict[str, int]:
    r"""
    Translate a glyphsizes.bin file into an "advances" mapping, which goes inside a "space" provider.

    .. note::

        This function does not return a new provider.

    >>> to_advances(b"\x00\x00\x00...")
    {
        "\x00": 0,
        "\x01": 0,
        "\x02": 0,
        "\uBEEF": 15
    }

    :param glyphsizes: The bytes of glyphsizes.bin.
    :param match_unicode_category:
        Only translate Unicode characters with these categories.
        By default, this is ``[Mc, Zp, Zs, Zl]``.
        This should cover most whitespace and marking characters.
    :returns: An "advances" dictionary.
    """
    if match_unicode_category is None:
        match_unicode_category = ["Mc", "Zp", "Zs", "Zl"]
    advances = {}
    for index, nibble in enumerate(glyphsizes):
        if unicodedata.category(character := chr(index)) in match_unicode_category:
            advances[character] = (nibble & 0xF) - (nibble >> 4 & 0xF)
    return advances


def build_template_for_character(
    character: str, template_storage: dict[int, PIL.Image.Image] | None = None
) -> PIL.Image.Image:
    """
    Build an image of a template :term:`codepoint`.

    Templates are simple white boxes with a codepoint :term:`texture` inside them.
    This is intended for use with a "legacy_unicode" :term:`provider`.

    :param character:
        A single character.
        This character's codepoint is what goes inside the box.
    :param template_storage:
        Dictionary of integers mapped to the template character PNG images.

        * `0-15` are images of hexadecimal digits.
        * `-4` and `-6` are "boxes" for 4 and 6 digit-long codepoints, respectively.

    :returns:
        A :class:`PIL.Image.Image` of the character template.
    """
    if template_storage is None:
        template_storage = TEMPLATE_CHARS
    codepoint = ord(character)
    if codepoint > 0xFFFF:
        box = template_storage[-6]
        # 0xABCDEF -> A
        box.paste(template_storage[(codepoint & 0xF00000) >> 20], (2, 2))
        # 0xABCDEF -> B
        box.paste(template_storage[(codepoint & 0xF0000) >> 16], (6, 2))
        # 0xABCDEF -> C
        box.paste(template_storage[(codepoint & 0xF000) >> 12], (10, 2))
        # 0xABCDEF -> D
        box.paste(template_storage[(codepoint & 0xF00) >> 8], (2, 9))
        # 0xABCDEF -> E
        box.paste(template_storage[(codepoint & 0xF0) >> 4], (6, 9))
        # 0xABCDEF -> F
        box.paste(template_storage[(codepoint & 0xF)], (10, 9))
    else:
        box = template_storage[-4]
        # 0xABCD -> A
        box.paste(template_storage[(codepoint & 0xF000) >> 12], (2, 2))
        # 0xABCD -> B
        box.paste(template_storage[(codepoint & 0xF00) >> 8], (6, 2))
        # 0xABCD -> C
        box.paste(template_storage[(codepoint & 0xF0) >> 4], (2, 9))
        # 0xABCD -> D
        box.paste(template_storage[codepoint & 0xF], (6, 9))
    return box


def align_unicode_page(sheet: PIL.Image.Image) -> PIL.Image.Image:
    """
    Align a Unicode page font sheet's characters to the left.

    This function is a shortcut for :func:`mcfonts.utils.resources.align_font_texture(sheet, (16, 16))`.

    :param sheet: The font sheet, not the individual character.
    :returns: The new font sheet.
    """
    return mcfonts.utils.resources.align_font_texture(sheet, (16, 16))



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""
Formats that can be passed into :meth:`mcfonts.MinecraftFont.export`.

All functions defined here must have a signature of::

   (glyphs: dict[str, mcfonts.glyphs.Glyph], settings: ExportSettings) -> bytes


"""
import io
import typing

import fontTools.ttLib
import lxml.etree

import mcfonts.glyphs
import mcfonts.utils.exporting.opentype
import mcfonts.utils.exporting.unihex
import mcfonts.utils.exporting.yaff
import mcfonts.utils.unihex


class ExportSettings(typing.TypedDict):
    """A TypedDict of the options that can be passed to any export format function."""

    name: str | None
    """The name of the resulting font, not what its filename is. Optional."""
    include_credits: bool | None
    """Whether to include basic copyright information and credits in the font file. Optional."""
    author: str | None
    """The author(s) of the font. Optional."""
    license: str | None
    """An instance of :class:`mcfonts.filters.Filter`. If supplied, filter rules will be obeyed. Optional."""


def opentype(glyphs: dict[str, mcfonts.glyphs.Glyph], settings: ExportSettings) -> bytes:
    """
    Export to an OpenType font with Type 2 Charstring outlines.

    The font is crafted through a TTX file (font XML), and characters are added in tables and
    given self-descriptive name mappings: ``("u0954", "u1fae4", "u2605")``.

    For some fields, the font's name will be "Minecraft <font name>".

    The font mustn't contain over 65,535 characters, or else any additional characters
    won't be added, and the font will be saved prematurely.

    :param glyphs: A dictionary of ``{character: glyph}``.
    :param settings: A dictionary conforming to :class:`ExportSettings`.
    :returns: An OpenType file binary.
    """
    font_xml = lxml.etree.XML(mcfonts.utils.exporting.opentype.XML_FONT_TEMPLATE)
    characters: set[str] = set()
    # Set the space character pre-emptively
    mcfonts.utils.exporting.opentype.set_space_to_character(font_xml, " ", 4)

    try:
        for character, glyph in glyphs.items():
            mcfonts.utils.exporting.opentype.include_glyph(font_xml, glyph, character, characters)
            characters.add(character)
    except mcfonts.exceptions.GlyphLimitError:
        pass

    if "name" in settings:
        font_name = settings["name"].encode("ascii", "ignore").decode("utf-8")  # Ignore anything not in ASCII.
    else:
        font_name = mcfonts.constants.UNKNOWN_FIELD
    mcfonts.utils.exporting.opentype.set_font_name(font_xml, font_name)

    if "author" in settings:
        mcfonts.utils.exporting.opentype.add_namerecord_to_font(font_xml, settings["author"], 9)

    if "include_credits" in settings:
        mcfonts.utils.exporting.opentype.add_namerecord_to_font(
            font_xml, f"mcfonts: {font_name} Regular: {hash(font_xml)}", 3
        )
        mcfonts.utils.exporting.opentype.add_namerecord_to_font(font_xml, "mcfonts", 8)
        mcfonts.utils.exporting.opentype.add_namerecord_to_font(
            font_xml, "https://gitlab.com/whoatemybutter/mcfonts", 11
        )
    if "license" in settings:
        mcfonts.utils.exporting.opentype.add_namerecord_to_font(font_xml, settings["license"], 0)
        mcfonts.utils.exporting.opentype.add_namerecord_to_font(font_xml, settings["license"], 7)
        mcfonts.utils.exporting.opentype.add_namerecord_to_font(font_xml, settings["license"], 13)
    else:
        mcfonts.utils.exporting.opentype.add_namerecord_to_font(font_xml, "Unknown license", 0)
        mcfonts.utils.exporting.opentype.add_namerecord_to_font(font_xml, "Unknown trademark.", 7)
        mcfonts.utils.exporting.opentype.add_namerecord_to_font(
            font_xml,
            (
                "This font is under no explicit licensing. "
                "This font's author may have additional licensing terms, contact them."
            ),
            13,
        )

    mcfonts.utils.exporting.opentype.set_ulunicoderange(font_xml, characters)
    mcfonts.utils.exporting.opentype.set_font_times(font_xml)

    sstr = io.BytesIO()
    with fontTools.ttLib.ttFont.TTFont(recalcTimestamp=False, recalcBBoxes=False, sfntVersion="OTTO") as font:
        font.importXML(io.StringIO(lxml.etree.tostring(font_xml, encoding=str)))
        font.save(sstr)
    return sstr.getvalue()


def yaff(glyphs: dict[str, mcfonts.glyphs.Glyph], settings: ExportSettings) -> bytes:
    """
    Export to a YAFF font file.

    For details on the YAFF format, see https://github.com/robhagemans/monobit/blob/master/YAFF.md.
    There is no limit to the number of characters.

    :param glyphs: A dictionary of ``{character: glyph}``.
    :param settings: A dictionary conforming to :class:`ExportSettings`.
    :returns: A YAFF file encoded in UTF-8 **bytes**.
    """
    store: dict[str, str | int | dict[str, list[str]]] = {
        "name": settings.get("name", mcfonts.constants.UNKNOWN_FIELD),
        "family": settings.get("name", mcfonts.constants.UNKNOWN_FIELD),
        "author": settings.get("author", "Unknown author"),
        "copyright": settings.get("license", "Unknown copyright"),
        "source-format": "minecraft-java-font",
        "encoding": "unicode",
        "default-char": "notdef",
        "word-boundary": "u+0020",
    }

    if not settings.get("license"):
        store["notice"] = (
            "This font is under no explicit licensing. "
            "This font's author may have additional licensing terms, contact them."
        )

    # Set the space character pre-emptively
    mcfonts.utils.exporting.yaff.set_space_to_character(store, " ", 4)
    for character, glyph in glyphs.items():
        mcfonts.utils.exporting.yaff.include_glyph(store, glyph, character)

    ret = []

    for key, value in store.items():
        if isinstance(value, str):
            ret.append(f"{key}: {value}")
        elif isinstance(value, dict):
            ret.append(f"{key}:")
            for subkey, subvalue in value.items():
                if subkey == mcfonts.utils.exporting.yaff.YAFF_STORE_GLYPH_MAGIC:
                    for line in value.get(mcfonts.utils.exporting.yaff.YAFF_STORE_GLYPH_MAGIC, []):
                        ret.append(f"\t{line}")
                else:
                    ret.append(f"\t{subkey}: {subvalue}")

    return "\n".join(ret).encode("utf-8")


def unihex(glyphs: dict[str, mcfonts.glyphs.Glyph], settings: ExportSettings) -> bytes:
    """
    Export to a GNU Unifont .hex format.

    For details on the Unifont .hex format, see https://en.wikipedia.org/wiki/GNU_Unifont#.hex_format.
    There is no limit to the number of characters.

    :param glyphs: A dictionary of ``{character: glyph}``.
    :param settings: A dictionary conforming to :class:`ExportSettings`. This is not used.
    :returns: A .hex file encoded in UTF-8 **bytes**.
    """
    store: dict[str, str] = {}
    mcfonts.utils.exporting.unihex.set_space_to_character(store, " ", 4)
    for character, glyph in glyphs.items():
        mcfonts.utils.exporting.unihex.include_glyph(store, glyph, character)
    ret = []
    meta = []
    if "name" in settings:
        meta.append(f"# Name: {settings['name']}")
    if "author" in settings:
        meta.append(f"# Author: {settings['author']}")
    if "license" in settings:
        meta.append(f"# License: {settings['license']}")
    if "include_credits" in settings:
        meta.append("# Generated by mcfonts: https://gitlab.com/whoatemybutter/mcfonts")

    for codepoint, data in store.items():
        ret.append(f"{codepoint}:{data}")

    return ("\n".join(meta) + "\n".join(sorted(ret, key=mcfonts.utils.unihex.get_codepoint))).encode("utf-8")



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""Typed dictionaries for providers in JSON form. See :func:`typing.TypedDict`."""

import typing

TypedProviderBitmap = typing.TypedDict(
    "TypedProviderBitmap",
    {"type": str, "file": str, "height": typing.NotRequired[int], "ascent": int, "chars": list[str]},
)

TypedProviderSpace = typing.TypedDict("TypedProviderSpace", {"type": str, "advances": dict[str, float | int]})

TypedProviderLegacyUnicode = typing.TypedDict(
    "TypedProviderLegacyUnicode", {"type": str, "template": str, "sizes": str}
)

TypedProviderReference = typing.TypedDict("TypedProviderReference", {"type": str, "id": str})

TypedProviderTTF = typing.TypedDict(
    "TypedProviderTTF",
    {
        "type": str,
        "file": str,
        "shift": list[float | int],
        "size": float | int,
        "oversample": float | int,
        "skip": str | list[str],
    },
)

TypedProviderUnihexSizeOverride = typing.TypedDict(
    "TypedProviderUnihexSizeOverride", {"from": str, "to": str, "left": int, "right": int}
)

TypedProviderUnihex = typing.TypedDict(
    "TypedProviderUnihex",
    {"type": str, "hex_file": str, "size_overrides": list[TypedProviderUnihexSizeOverride]},
)

TypedProviderDict = typing.Union[
    TypedProviderUnihex,
    TypedProviderTTF,
    TypedProviderSpace,
    TypedProviderBitmap,
    TypedProviderLegacyUnicode,
    TypedProviderReference,
]

TypedProvidersDict = typing.TypedDict("TypedProvidersDict", {"providers": list[TypedProviderDict]})



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""Utilities for working with paths, resources, and loading resources."""
from __future__ import annotations

import collections.abc
import itertools
import os
import pathlib
import re
import zipfile

import PIL.Image
import PIL.ImageChops

import mcfonts.utils.image


def resolve_path(resource: str, base: pathlib.Path, subpath: str | None = None) -> pathlib.Path:
    """
    Resolve the correct path to the resource indicated by `path`, using `origin` as the base path to follow from.

    `origin` is where the original calling resource is;
    ``path`` is the requested resource path from the requesting resource.

    .. warning::

        This function doesn't return extensions.
        You must add those later.

    >>> resolve_path("mypath", pathlib.Path("~/mcfonts"), "textures")
    '~/mcfonts/textures/mypath'

    :param resource: An unresolved path to the resource.
    :param base: A real path to the central ``/assets/`` directory.
    :param subpath: Under what subdirectory in `base` to start looking for a resource. Optional.
    :returns: An absolute path of the requested resource.
    """
    temppath = str(resource).split(":", 1)
    if len(temppath) > 1:
        path_suffix = f"{temppath[0]}/{subpath}/{temppath[1]}" if subpath else f"{temppath[0]}/{temppath[1]}"
    else:
        path_suffix = f"minecraft/{subpath}/{temppath[0]}" if subpath else f"minecraft/{temppath[0]}"
    return expand_path(base.joinpath(path_suffix))


def expand_path(path: os.PathLike[str] | str) -> pathlib.Path:
    """
    Expand `path` with any variables or shortcuts, such as ``~``, ``$HOME``, etc.

    >>> expand_path("~/Documents")
    '/home/me/Documents'

    :param path: The unexpanded path.
    :returns: The expanded absolute path.
    """
    return pathlib.Path(path).expanduser()


def expand_resource_location(file_path: pathlib.Path) -> str:
    """
    Take any path to a texture file and return the Minecraft file string for it.

    :param file_path: A real path to any file.
    :returns: A string in the format of "<namespace>?:<dir/>?<file>".
    """
    splitted = re.split(f"[{os.path.sep}:]", str(file_path))
    try:
        assets = splitted.index("assets")
    except ValueError:
        assets = -1
    if (namespace := splitted[0]) == "minecraft":
        return f"{splitted[assets + 3]}/{'/'.join(splitted[assets + 4:])}"
    return f"{namespace}/{splitted[assets + 3]}/{'/'.join(splitted[assets + 4:])}"


def traverse_to_assets(path: pathlib.Path) -> pathlib.Path:
    """
    Traverse up the path until the ``/assets/`` directory is found, and then cut off the path there.

    :param path: The path to the file.
    :returns:
        The cut-off path up to the ``/assets/`` directory if found.
    :raises FileNotFoundError: If the ``/assets/`` directory could not be found.
    """
    current_path = path.absolute()

    n = 500
    while n > 0:
        if current_path.name == "assets":
            return current_path
        current_path = current_path.parent
        n -= 1
    raise FileNotFoundError


def divide_resource_by_grid(
    resource: PIL.Image.Image, row_column_count: tuple[int, int]
) -> collections.abc.Iterable[PIL.Image.Image | None]:
    """
    Yield images of the sections of a divided resource.

    Given a resource and number of characters in each row and column of that resource, yield every
    :class:`PIL.Image.Image` in that resource.

    If no data exists in a cell, yield None.

    :param resource: A :class:`PIL.Image.Image` of the font resource.
    :param row_column_count: A tuple of the number of characters in each row and column of `resource`.
    :returns: A yield of :class:`PIL.Image.Image` glyphs, or None.
    """
    height = resource.height // row_column_count[1]
    width = resource.width // row_column_count[0]
    for y_index in range(row_column_count[1]):
        for x_index in range(row_column_count[0]):
            if not mcfonts.utils.image.is_image_empty(
                glyph := resource.crop(
                    (
                        x_index * width,
                        y_index * height,
                        (x_index + 1) * width,
                        (y_index + 1) * height,
                    )
                )
            ):
                yield glyph
            else:
                yield None


def load_bitmap(file_name: str, base: pathlib.Path) -> PIL.Image.Image:
    """
    Companion function to :func:`load_resources` but specialized for bitmap provider resources.

    :param file_name: The name of the file to search for.
    :param base: The path to the base "assets" directory.
    :return: A :class:`PIL.Image.Image` instance.
    """
    with open(
        resolve_path(file_name, base, "textures"),
        "rb",
    ) as open_tempfile:
        image = PIL.Image.open(open_tempfile, formats=["png"])
        image.load()
    return image


def load_resources_legacy_unicode(
    template: str,
    sizes: str,
    base: pathlib.Path,
) -> collections.abc.Iterable[tuple[str, PIL.Image.Image | bytes]]:
    """
    Companion function to :func:`load_resources` but specialized for legacy_unicode provider resources.

    :param template: The template string, which must include "%s".
    :param sizes: The name of the "sizes" file.
    :param base: The real path to the base ``/assets/`` directory.
    :returns: A yield of {filename: image}.
    """
    template_parts: list[str] = template.split("%s", 1)
    # Skip surrogates
    for codepage in itertools.chain(range(0xD8), range(0xE0, 0x100)):
        template_name = f"{template_parts[0]}{codepage:02x}{template_parts[1]}"
        try:
            with open(
                resolve_path(template_name, base, "font"),
                "rb",
            ) as open_tempfile:
                image = PIL.Image.open(open_tempfile, formats=["png"])
                image.load()
                yield template_name, image
        except FileNotFoundError:
            # Page doesn't exist, acceptable error
            continue
    with open(
        resolve_path(sizes, base, "font"),
        "rb",
    ) as open_tempfile:
        yield sizes, open_tempfile.read()


def load_resources_unihex(file_name: str, base: pathlib.Path) -> collections.abc.Iterable[tuple[str, str]]:
    """
    Companion function to :func:`load_resources` but specialized for unihex provider resources.

    :param file_name: The path to the .zip file that contains the .hex files.
    :param base: The real path to the base ``/assets/`` directory.
    :returns: Yield of ``(filename, contents)``.
    """
    with open(resolve_path(file_name, base), "rb") as hex_zip:
        with zipfile.ZipFile(hex_zip) as memory_zip:
            for filename in memory_zip.namelist():
                if filename.lower().endswith(".hex"):  # Only care about .hex files.
                    yield filename, memory_zip.read(filename).decode("utf-8")


def align_font_texture(
    sheet: PIL.Image.Image, glyph_cell: tuple[int, int] = (8, 8), new_left_padding: int = 0
) -> PIL.Image.Image:
    """
    Align a font sheet's character's sheet to what `new_left_padding` is.

    This only shifts the character on the X axis (horizontally, left-right).

    :param sheet: The font sheet, not the individual glyph.
    :param glyph_cell: The dimensions of each individual character, their cell bounding box.
    :param new_left_padding: The distance away from the left edge the new character should be.
    :returns: The new font sheet.
    """
    height = sheet.height // glyph_cell[1]
    width = sheet.width // glyph_cell[0]
    for y_position in range(height):
        for x_position in range(width):
            # Glyph dimensions before alignment.
            dimensions = (
                x_position * glyph_cell[0],
                y_position * glyph_cell[1],
                (x_position + 1) * glyph_cell[0],
                (y_position + 1) * glyph_cell[1],
            )
            # Get the glyph itself.
            glyph = sheet.crop(dimensions)
            # If width is not the new padding, offset it by the difference.
            if (margin := mcfonts.utils.image.get_image_bearings(glyph)[0]) != new_left_padding:
                sheet.paste(PIL.ImageChops.offset(glyph, new_left_padding - margin, 0))
    return sheet



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""
Coverage reports are instances of :class:`mcfonts.coverage_reports.CoverageReport`.

They contain information on the :term:`coverage` of a specific font,
and contain functions for displaying this information.
"""
from __future__ import annotations

import typing
import unicodedata

import tinyunicodeblock

import mcfonts.utils.colors

PRIVATE_USE_PREFIX: str = "Private Use"


class CoverageReport:
    """
    A coverage report.

    Contains information of the number of characters covered, and the blocks covered.
    """

    def __init__(self, chars: set[str], blocks: dict[str, int]) -> None:
        """
        Construct a coverage report.

        :param chars: A set of individual characters.
        :param blocks: A dictionary of ``{block name: chars covered}``.
        """
        self.chars: set[str] = chars
        """
        A set of the characters.
        """
        self.blocks: dict[str, int] = blocks
        """
        A dictionary of ``{block name: number of characters of this block}``.
        """

    def block_progress(self, show_empty: bool = True) -> None:
        """
        Print the block progress in each block.

            ✓ Basic Latin (U+0000-U+007F) [95/95, 100.00%]
            ~ Latin-1 Supplement (U+0080-U+00FF) [94/96, 97.92%]

        And, if `show_empty`:

            ❌ Arabic (U+0600-U+06FF) [0/255, 0.00%]

        :param show_empty:
            If True, blocks with no support will be shown.
            If False, blocks with no support will be omitted.
        """
        for block in tinyunicodeblock.BLOCKS_BYNAME:
            info = tinyunicodeblock.BLOCKS_BYNAME[block]
            start = ord(info[0])
            end = ord(info[1])
            number = self.blocks.get(block, 0)
            amount = end - start - info[2] + 1
            if block in self.blocks:
                if number >= amount:
                    if mcfonts.utils.colors.USE_COLORS:
                        print(f"{mcfonts.utils.colors.GREEN_FORE}✓ {block}{mcfonts.utils.colors.RESET_FORE}", end=" ")
                    else:
                        print(f"✓ {block}", end=" ")
                else:
                    if mcfonts.utils.colors.USE_COLORS:
                        print(f"{mcfonts.utils.colors.YELLOW_FORE}~ {block}{mcfonts.utils.colors.RESET_FORE}", end=" ")
                    else:
                        print(f"~ {block}", end=" ")
            elif show_empty:
                if mcfonts.utils.colors.USE_COLORS:
                    print(f"{mcfonts.utils.colors.RED_FORE}❌ {block}{mcfonts.utils.colors.RESET_FORE}", end=" ")
                else:
                    print(f"❌ {block}", end=" ")
            try:
                print(f"(U+{start:04X}-U+{end:04X}) [{number}/{amount}, {number / amount * 100:.2f}%]")
            except ZeroDivisionError:
                print(f"(U+{start:04X}-U+{end:04X}) [{number}, {number}.00%]")

    def compare(self, other: typing.Self) -> None:
        """
        Given `other`, a second instance of :class:`CoverageReport`, compare the two, using `self` as a baseline.

        The information compared is:

        * Character count
        * Blocks covered

        :param other: A second instance of :class:`mcfonts.coverage_reports.CoverageReport` to compare to.
        """
        if mcfonts.utils.colors.USE_COLORS:
            print(f"{mcfonts.utils.colors.BRIGHT}COUNTS{mcfonts.utils.colors.RESET_ALL}")
        else:
            print("COUNTS")
        print(":: stat: this | other (delta)")
        len_self = len(self.chars)
        len_other = len(other.chars)
        print(
            f"Characters: {len_self:,} | {len_other:,} " f"({mcfonts.utils.colors.color_number(len_other - len_self)})"
        )
        len_blocks = (len(self.blocks), len(other.blocks))
        print(
            f"Blocks: {len_blocks[0]:,} | {len_blocks[1]:,} ("
            f"{mcfonts.utils.colors.color_number(len_blocks[1] - len_blocks[0])})\n"
        )
        if mcfonts.utils.colors.USE_COLORS:
            print(f"{mcfonts.utils.colors.BRIGHT}BLOCKS{mcfonts.utils.colors.RESET_ALL}")
        else:
            print("BLOCKS")
        print(":: block: this | other (delta)")
        for block_name in (self.blocks | other.blocks).keys():
            block_info: tuple[int, int] = tinyunicodeblock.BLOCKS_BYNAME.get(block_name, (0, 0))
            block_this = self.blocks.get(block_name) or 0
            block_other = other.blocks.get(block_name) or 0
            print(f"{chr(block_info[0] + 14)} ", end="")
            if mcfonts.utils.colors.USE_COLORS:
                print(f"{mcfonts.utils.colors.BRIGHT}{block_name}{mcfonts.utils.colors.RESET_ALL}")
            else:
                print(block_name)
            print(
                f"\t{block_this:,}/{block_info[1]:,} | {block_other:,}/{block_info[1]:,} "
                f"({mcfonts.utils.colors.color_number(block_other - block_this)})"
            )

    def character_table(self, show_offsets: bool = True, unknown_character: str = "▯") -> None:
        """
        Print a table of every character in this font.

        These are organized by their block, and order of appearance in the providers used to declare them.

        If :data:`mcfonts.colors.USE_COLORS` is True, a red or green color will be used to indicate if the character
        exists. If False, `padding_character` will be displayed if the character does not exist.

        .. important::

            If a character is Private Use, it will not be shown.

        Example::

                ▼ Block Elements (U+2580-U+259F) ▼
                   0 1 2 3 4 5 6 7 8 9 A B C D E F
            U+258X ▀ ▁ ▂ ▃ ▄ ▅ ▆ ▇ █ ▉ ▊ ▋ ▌ ▍ ▎ ▏
            U+259X ▐ ░ ▒ ▓ ▔ ▕ ▖ ▗ ▘ ▙ ▚ ▛ ▜ ▝ ▞ ▟

        :param show_offsets:
            If True, show the individual offsets for each character in a table-like format.
            If False, don't show offsets.
        :param unknown_character:
            The character to display when a font has no glyph for a character in a table.
            The default is ``U+25AF ▯ WHITE VERTICAL RECTANGLE``.
        """
        for block in self.blocks:
            if PRIVATE_USE_PREFIX in block:
                # Skip PUA blocks
                continue
            info = tinyunicodeblock.BLOCKS_BYNAME.get(block, ("\0", "\0", 0))
            start = ord(info[0])
            end = ord(info[1])

            print(f"\n\t▼ {block} (U+{start:04X}-U+{end:04X}) ▼")
            if show_offsets:
                print("0 1 2 3 4 5 6 7 8 9 A B C D E F".rjust(len(f"{end:04X}") + 34))

            for row in range((end - start + 1) // 16):
                row_offset = start + (row * 16)
                if show_offsets:
                    print(f"U+{row_offset // 16:04X}X", end=" ")
                for codepoint in range(row_offset, row_offset + 16):
                    if (character := chr(codepoint)) in self.chars:
                        if mcfonts.utils.colors.USE_COLORS:
                            print(
                                f"{mcfonts.utils.colors.GREEN_FORE}{character}{mcfonts.utils.colors.RESET_FORE}",
                                end=" ",
                            )
                        else:
                            print(character, end=" ")
                    else:
                        if unicodedata.name(character, False) and mcfonts.utils.colors.USE_COLORS:
                            print(
                                f"{mcfonts.utils.colors.RED_FORE}{character}{mcfonts.utils.colors.RESET_FORE}", end=" "
                            )
                        else:
                            print(unknown_character, end=" ")
                print("")

    def block_summary(self, show_empty_blocks: bool = False) -> None:
        """
        Print a padded list of every block in the font, and their completion status.

            0000-007F     Basic Latin          Complete
            0080-00FF     Latin-1 Supplement   Partial

        And, if `show_empty_blocks` is True:

            0600-06FF     Arabic               None

        :param show_empty_blocks:
            If True, blocks with no support will be shown.
            If False, blocks with no support will be omitted.
        """
        block_name_length = 1
        # Have to know in advance the padding for block names
        for block_name in tinyunicodeblock.BLOCKS_BYNAME | tinyunicodeblock.CSUR_BYNAME:
            if block_name in self.blocks and (len_block := len(block_name)) > block_name_length:
                block_name_length = len_block
        for block, info in tinyunicodeblock.BLOCKS_BYNAME.items():
            # Use BLOCKS_BYNAME to support empty blocks.
            new_block = block
            if not show_empty_blocks and new_block not in self.blocks:
                continue
            if new_block in tinyunicodeblock.CSUR_BYNAME and mcfonts.utils.colors.USE_COLORS:
                new_block = f"{mcfonts.utils.colors.MAGENTA_BACK}{mcfonts.utils.colors.BLACK_FORE}{new_block}"
            start = ord(info[0])
            end = ord(info[1])
            block_range = f"{start:04X}-{end:04X}".ljust(13)
            amount = self.blocks.get(new_block, 0)
            if amount >= (end - start - info[2] + 1):
                completion = (
                    f"{mcfonts.utils.colors.GREEN_BACK}{mcfonts.utils.colors.BLACK_FORE}Complete"
                    if mcfonts.utils.colors.USE_COLORS
                    else "Complete"
                )
            elif amount > 0:
                completion = (
                    f"{mcfonts.utils.colors.YELLOW_BACK}{mcfonts.utils.colors.BLACK_FORE}Partial"
                    if mcfonts.utils.colors.USE_COLORS
                    else "Partial"
                )
            else:
                completion = (
                    f"{mcfonts.utils.colors.RED_BACK}{mcfonts.utils.colors.BLACK_FORE}None"
                    if mcfonts.utils.colors.USE_COLORS
                    else "None"
                )
            if mcfonts.utils.colors.USE_COLORS:
                print(
                    f"{block_range} {new_block.ljust(block_name_length + 1)} "
                    f"{completion}{mcfonts.utils.colors.RESET_BACK}{mcfonts.utils.colors.RESET_FORE}"
                )
            else:
                print(f"{block_range} {new_block.ljust(block_name_length + 1)} {completion}")



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""General exceptions."""
from __future__ import annotations


class ExportError(Exception):
    """Base parent class for all exceptions that may be raised when exporting a provider's characters."""


class GlyphLimitError(Exception):
    """Raised when a font has 65,535 characters allocated and that no more can be added."""


class ProviderError(Exception):
    """Raised when a :class:`mcfonts.providers.base.Provider` is expected but not found."""


class ReferenceChildError(Exception):
    """Raised when another reference provider is given as the child provider for a reference provider."""


class ResourceError(ExportError):
    """Raised when a resource is expected but does not exist."""


class CharacterPaddingError(ExportError):
    """Raised when a requested character is in :data:`mcfonts.constants.PADDING_CHARS`."""



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""A ttf provider, including already-made TTF and OTF fonts."""
import collections.abc
import io
import os

import fontTools.ttLib

import mcfonts.providers.base
import mcfonts.utils.schemas


class TTFProvider(mcfonts.providers.base.Provider):
    """
    The ``ttf`` provider embeds already-compiled TrueType and OpenType fonts.

    mcfonts does not handle TTF providers by default, and they cannot be exported.
    """

    def __init__(
        self,
        provider: mcfonts.utils.types.TypedProviderTTF,
        origin: os.PathLike[str] | str,
        resource: bytes | None = None,
    ) -> None:
        self.contents = provider
        super().__init__("ttf", origin)
        if "skip" in self.contents:
            chars = []
            if isinstance(skip := self.get_contents().get("skip"), list):
                for line in skip:
                    chars.append(line)
            # Flatten into a string
            self.contents["skip"] = "".join(chars)
        self.validate()
        self.resource = resource
        if self.resource:
            self.font = fontTools.ttLib.TTFont(io.BytesIO(self.resource))
        else:
            self.font = None

    def pretty_print(self) -> str:
        return f"{self.get_icon()} ttf: {self.get_contents().get('file', 'no file')}"

    def print_info(self) -> None:
        super().print_info()
        if self.font:
            print(f"Characters: {self.font['maxp'].numGlyphs:,}")

    def construct_glyphs(self) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.Glyph]]:
        yield from []

    def yield_glyphs(self) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.Glyph]]:
        yield from []

    def yield_characters(self) -> collections.abc.Iterable[str]:
        yield from []

    def get_icon(self) -> str:
        return "𝐓"

    def get_pack_versions(self) -> tuple[int, None]:
        return 9, None

    def yield_glyphs_in_unirange(
        self, unirange_notation: str
    ) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.BitmapGlyph]]:
        yield from []

    def get_contents(self) -> mcfonts.utils.types.TypedProviderTTF:
        return self.contents
